/*-
 * Copyright (c) 2008 Jean-Yves Moulin <jym@ldh.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/malloc.h>
#include <sys/time.h>
#include <sys/systm.h>
#include <sys/ioccom.h>
#include <sys/conf.h>
#include <sys/mutex.h>
#include <sys/syslog.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>

#ifdef __NetBSD__
#include <sys/cdefs.h>
#include <machine/param.h>
#include <sys/lkm.h>
#include <machine/limits.h>
#if (__NetBSD_Version__ >= 399002000)
# include <sys/kauth.h>
#endif
#else /* FreeBSD */
#include <sys/module.h>
#include <sys/libkern.h>
#include <machine/float.h>
#endif

#include <net/route.h>
#include <sys/mbuf.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <net/pfil.h>

#include "lbdr.h"


/***********
 * MACRO
 */


/*  Into Makefile:
 * #define NOBIG 1
 * #define DEBUG 1
 * #define DEBUGTIME 1
 * #define DEGROBUG 1
 * #define DEBUGTRACE 1
 * #define DEBUGSCHED 1
 */

/* Sorry for this macro but I'm borring with #ifdef ... #endif */
#ifdef DEBUG
#  define D(X) X
#  ifdef DEGROBUG
#     define DG(X) X
#  endif
#  ifdef DEBUGTRACE
#    define DT(X) X
#  endif
#  ifdef DEBUGSCHED
#     define DS(X) X
#  endif
/* Debug only, I need to know where time is lost (it's real time and
 * not kernel time)
 */
#  ifdef DEBUGTIME
#    define INIT(X, Y)	X.ns = 0 ; X.nb = 0;			\
			F(if(!mtx_initialized(&(X.mtx))))  	\
				mtx_init(&(X.mtx), Y, NULL, MTX_DEF);

#    define START(X)	mtx_lock(&(X.mtx));			\
			getnanotime(&(X.deb));			\
			X.nb++;

#    define STOP(X)	getnanotime(&(X.fin));			\
			X.ns += ((X.fin.tv_sec  - X.deb.tv_sec) * 1000000000 + \
			(X.fin.tv_nsec - X.deb.tv_nsec));	\
			mtx_unlock(&(X.mtx));
#  endif /* DEBUGTIME */

#define DEINIT(X)	mtx_destroy(&(X.mtx));

#define LOG(X, ...) {   mtx_lock(&mtx_log);			\
			log(X, __VA_ARGS__);			\
			mtx_unlock(&mtx_log); }

#endif /* DEBUG */

#ifndef D
# define D(X)
#endif
#ifndef DG
# define DG(X)
#endif
#ifndef DT
# define DT(X)
#endif
#ifndef DS
# define DS(X)
#endif
#ifndef INIT
# define INIT(X, Y)
#endif
#ifndef START
# define START(X)
#endif
#ifndef STOP
# define STOP(X)
#endif
#ifndef DEINIT
# define DEINIT(X)
#endif
#ifndef LOG
#  define LOG(X, ...)
#endif



#define FOREACHCLONE(X, ...) {					\
		struct lbdr_service *clone;			\
		struct lbdr_service *nextclone = X->clone;	\
		while(nextclone) {				\
			clone = nextclone;			\
			nextclone = clone->clone;		\
			__VA_ARGS__;				\
		} }

#define SPORT(ip)	ip->ss_family == AF_INET ? 		\
			((struct sockaddr_in *)ip)->sin_port :	\
			((struct sockaddr_in6 *)ip)->sin6_port

 
#ifndef NOBIG
#  define MPSAFE		1	/* not enough to avoid race condition */
#  define G(X)			X
#  define BIG_LOCK()		mtx_lock(&big_lock)
#  define BIG_UNLOCK()	mtx_unlock(&big_lock)
#else
#  define MPSAFE		1
#  define G(X)
#  define BIG_LOCK()
#  define BIG_UNLOCK()
#endif



/***
 * Configurable parameters:
 */
/* hash table size **** BE CAREFULL: must be a power of 2 **** */
#define HT_SERVICE      128
#define MAX_SERVER    32767	/* 2^15 */
#define LOGLIMIT	500	/* print syslog every 500 occurences */

/* parametres par defaut: */
#define TIMEOUT_SYN	  5	/* secondes avant de passer en inactive */
#define TIMEOUT_ESTABL	 60	/* secondes avant de passer en inactive */
#define TIMEOUT_CLOSE	120	/* cnx inactive, sec avant de virer de la HT */

/* enum cnx_status { CNX_SYN, CNX_ESTABL, CNX_TIMEOUT, CNX_CLOSE }; */
#define CNX_FLAG_EXIST  0x01
#define CNX_FLAG_ACT    0x02
#define CNX_FLAG_DATA   0x04
#define CNX_SYN         0x03 /*  CNX_FLAG_ACT  && !CNX_FLAG_DATA  &&  CNX_FLAG_EXIST */
#define CNX_ESTABL      0x07 /*  CNX_FLAG_ACT  &&  CNX_FLAG_DATA  &&  CNX_FLAG_EXIST */
#define CNX_TIMEOUT     0x01 /* !CNX_FLAG_ACT  && !CNX_FLAG_DATA  &&  CNX_FLAG_EXIST */
#define CNX_CLOSE       0x05 /* !CNX_FLAG_ACT  &&  CNX_FLAG_DATA  &&  CNX_FLAG_EXIST */

/* enum srv_status { SRV_UP, SRV_DOWN }; */
#define SRV_DOWN        0x00
#define SRV_UP          0x01

#define LBDR_VERSION "0.7 IPv6 ("__DATE__" "__TIME__")"

/* enum mbuf_status { NOTHING, DROP }; */
#define NOTHING		0x01
#define DROP		0x02

/****
 * OS dependencies
 */

#ifdef __NetBSD__
/* #warning "*** NetBSD version currently unusable (problem with rtalloc1()) ***" */
# define N(X) X
# define F(X)
# define mtx_lock(X)		mutex_enter(X)
# define mtx_init(X, Y, Z, W)	mutex_init(X, MUTEX_DEFAULT, IPL_NONE)
# define mtx_unlock(X)		mutex_exit(X)
# define mtx_destroy(X)		mutex_destroy(X)
# define callout_deactivate(X)	callout_ack(X)
# define rtalloc1(X, Y, Z)	rtalloc1(X, 0);
/* TODO: not a very good solution ... : */
# define callout_drain(X)	if(!callout_pending(X)) \
					callout_stop(X); \
				callout_destroy(X);


int drlb_lkmentry(struct lkm_table *, int, int);

#else /* __FreeBSD__ */

# define F(X) X
# define N(X)

static struct cdev *sdev;
int drlb_ioctl(struct cdev *dev, u_long cmd, caddr_t arg, int mode, struct thread *td);
static struct cdevsw drlb_devsw = { 
	.d_version = D_VERSION,
	.d_ioctl = drlb_ioctl,
	.d_name = "drlb"
};
typedef struct mtx kmutex_t;

#endif


/*************
 * DATA types
 */

union protohdr {
	unsigned char *raw;
	struct udphdr *uh;
	struct tcphdr *th;
	struct icmp *icmph;
	u_int16_t *portp;
};

struct lbdr_server {
        struct rtentry *rt;
        struct sockaddr_storage *ip;
        uint32_t active_cnx;
        uint32_t inactive_cnx;
        uint32_t weight:10,
		 status:1,
        	 data:20;
	uint32_t nb_cnx;
        uint32_t pkts;
        uint32_t bytes;
        uint32_t no_fin;
        uint32_t err_unreach;
};

struct lbdr_conn {			/* 164o*/
	struct sockaddr_storage ip_src;	/* 128 */
	struct callout cal;		/*  16 */
	struct lbdr_server *srv;	/*   4 */
	struct lbdr_conn *parent;	/*   4 */
	struct lbdr_conn *next;		/*   4 */ /* sll puisque risque de collisions dans HT */
	uint32_t hash:29,		/*   4 */
		 status:3;		/*   . */
	kmutex_t *pmtx;			/*   4 */
};


struct counters {
	uint32_t t_syn;
	uint32_t t_est;
	uint32_t nb_coll;
	uint32_t err_nosrv;
	uint32_t err_nosyn;
};

struct lbdr_service {
        struct sockaddr_storage *ip;
        struct lbdr_server **server;
        struct lbdr_conn *htable;
        struct lbdr_service *next;		/* sll puisque risque de collisions dans HT */
	struct lbdr_service *clone;
	struct counters *to;
	int (*sched)(struct lbdr_server **, int, int);
	kmutex_t *locktable;
	uint32_t htsize;
	uint32_t refcnt;
	uint32_t nb_server:15,
		 used_server:15,
		 cloned:1,
		 tls:1;
};


/*************
 * DATA
 */

static int kld_loaded = 0;
static int nticks = 0;

#define IP_BUF 46 /* IPv6 == 39 + ':' + port (max 5char) */
static char ip_buf[IP_BUF];

static struct lbdr_service **service_available = NULL;
static struct lbdr_info info;

static uint32_t t_clo;
G(static kmutex_t big_lock;)
D(static kmutex_t mtx_log;)

#ifdef DEBUG
struct timelost {
	struct timespec deb;
	struct timespec fin;
	unsigned int nb;
	long long unsigned int ns;
	kmutex_t mtx;
};
static struct timelost inp_fwd, inp_new, inp_old, fd_svi, fd_cnx, dr_cnx, cl_cnx, cr_cnx, schd;
#endif

static MALLOC_DEFINE(M_BUF, "DRLB", "DRLB Memory Buffers");



/*************
 * misc functions
 */
static __inline__ uint32_t sfhash (struct sockaddr_storage *ip)
{
	struct sockaddr_in *in4 = (struct sockaddr_in *)ip;
	struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)ip;
	register uint32_t hash, tmp;

	if(ip->ss_family == AF_INET6) {
		hash   = (in6->sin6_addr.s6_addr16[0]) ;
		tmp    = (in6->sin6_addr.s6_addr16[1] << 11) ^ hash;
		hash   = (hash << 16) ^ tmp;
		hash  += hash >> 11;
		hash  += in6->sin6_addr.s6_addr16[2];
		tmp    = (in6->sin6_addr.s6_addr16[3] << 11) ^ hash;
		hash   = (hash << 16) ^ tmp;
		hash  += hash >> 11;
		hash  += in6->sin6_addr.s6_addr16[4];
		tmp    = (in6->sin6_addr.s6_addr16[5] << 11) ^ hash;
		hash   = (hash << 16) ^ tmp;
		hash  += hash >> 11;
		hash  += in6->sin6_addr.s6_addr16[6];
		tmp    = (in6->sin6_addr.s6_addr16[7] << 11) ^ hash;
	} else {
		hash  = (in4->sin_addr.s_addr & 0x0000ffff) ;
		tmp   = (((in4->sin_addr.s_addr & 0xffff0000) >> 16) << 11) ^ hash;
	}

	hash  = (hash << 16) ^ tmp;
	hash += hash >> 11;
	hash += ntohs(SPORT(ip));
	hash ^= hash << 11;
	hash += hash >> 17;
	hash ^= hash << 3;
	hash += hash >> 5;
	hash ^= hash << 4;
	hash += hash >> 17;
	hash ^= hash << 25;
	hash += hash >> 6;

	return hash;
}




/**
 * print IP address and port, v4 or v6
 * DO NOT USE TWICE IN A PRINTF (because buffer is static)
 */
static char *paddr(struct sockaddr_storage *ip)
{
	struct sockaddr_in *in = (struct sockaddr_in *)ip;
	struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)ip;

	bzero(ip_buf, IP_BUF);

	if(ip->ss_family == AF_INET) {
		snprintf(ip_buf, IP_BUF, "%u.%u.%u.%u:%i",
			((unsigned char *)&in->sin_addr.s_addr)[0],
			((unsigned char *)&in->sin_addr.s_addr)[1],
			((unsigned char *)&in->sin_addr.s_addr)[2],
			((unsigned char *)&in->sin_addr.s_addr)[3],
			ntohs(in->sin_port));

	} else {
		snprintf(ip_buf, IP_BUF, "%x%x:%x%x:%x%x:%x%x:%x%x:%x%x:%x%x:%x%x.%i",
			((unsigned char *)&in6->sin6_addr.s6_addr)[0],
			((unsigned char *)&in6->sin6_addr.s6_addr)[1],
			((unsigned char *)&in6->sin6_addr.s6_addr)[2],
			((unsigned char *)&in6->sin6_addr.s6_addr)[3],
			((unsigned char *)&in6->sin6_addr.s6_addr)[4],
			((unsigned char *)&in6->sin6_addr.s6_addr)[5],
			((unsigned char *)&in6->sin6_addr.s6_addr)[6],
			((unsigned char *)&in6->sin6_addr.s6_addr)[7],
			((unsigned char *)&in6->sin6_addr.s6_addr)[8],
			((unsigned char *)&in6->sin6_addr.s6_addr)[9],
			((unsigned char *)&in6->sin6_addr.s6_addr)[10],
			((unsigned char *)&in6->sin6_addr.s6_addr)[11],
			((unsigned char *)&in6->sin6_addr.s6_addr)[12],
			((unsigned char *)&in6->sin6_addr.s6_addr)[13],
			((unsigned char *)&in6->sin6_addr.s6_addr)[14],
			((unsigned char *)&in6->sin6_addr.s6_addr)[15],
			ntohs(in6->sin6_port));
	}

	return ip_buf;
}


#ifdef DEBUG

static __inline__ const char *proto_name(unsigned proto)
{
        static char buf[20];
        switch (proto) {
                case IPPROTO_IP:
                        return "IP";
                case IPPROTO_UDP:
                        return "UDP";
                case IPPROTO_TCP:
                        return "TCP";
                case IPPROTO_ICMP:
                        return "ICMP";
                default:
                        sprintf(buf, "IP_%d", proto);
			return buf;
	}
	return "***";
}
#endif

/**
 * Is this better than memcmp() ?
 */
static __inline__ int same_addr(struct sockaddr_storage *a, struct sockaddr_storage *b)
{
	struct sockaddr_in *in4a, *in4b;
	struct sockaddr_in6 *in6a, *in6b;

	if(a->ss_family != b->ss_family) return 0;

        switch (a->ss_family) {
        case AF_INET:
		in4a = (struct sockaddr_in *)a;
		in4b = (struct sockaddr_in *)b;
                if( in4a->sin_addr.s_addr == in4b->sin_addr.s_addr &&
                           in4a->sin_port == in4b->sin_port)
			 return 1;
                break;
        case AF_INET6:
        	in6a = (struct sockaddr_in6 *)a;
        	in6b = (struct sockaddr_in6 *)b;

                if( in6a->sin6_addr.s6_addr32[0] == in6b->sin6_addr.s6_addr32[0] &&
                    in6a->sin6_addr.s6_addr32[1] == in6b->sin6_addr.s6_addr32[1] &&
                    in6a->sin6_addr.s6_addr32[2] == in6b->sin6_addr.s6_addr32[2] &&
                    in6a->sin6_addr.s6_addr32[3] == in6b->sin6_addr.s6_addr32[3] &&
			         in6a->sin6_port == in6b->sin6_port)
			return 1;
        	break;
        }

	return 0;
}



/**
 * I do not use memcpy() because I want all data to be fill (len, family..)
 */
static __inline__ void copy_addr(struct sockaddr_storage *dst, struct sockaddr_storage *src)
{
	struct sockaddr_in *dst4, *src4;
	struct sockaddr_in6 *dst6, *src6;

        switch (src->ss_family) {
        case AF_INET:
		dst4 = (struct sockaddr_in *)dst;
		src4 = (struct sockaddr_in *)src;

		dst4->sin_family = AF_INET;
		dst4->sin_len = sizeof(struct sockaddr_in);
                dst4->sin_addr.s_addr = src4->sin_addr.s_addr;
		dst4->sin_port = src4->sin_port;
        	return;

        case AF_INET6:
        	dst6 = (struct sockaddr_in6 *)dst;
        	src6 = (struct sockaddr_in6 *)src;

		dst6->sin6_family = AF_INET6;
		dst6->sin6_len = sizeof(struct sockaddr_in6);
                dst6->sin6_addr.s6_addr32[0] = src6->sin6_addr.s6_addr32[0];
                dst6->sin6_addr.s6_addr32[1] = src6->sin6_addr.s6_addr32[1];
                dst6->sin6_addr.s6_addr32[2] = src6->sin6_addr.s6_addr32[2];
                dst6->sin6_addr.s6_addr32[3] = src6->sin6_addr.s6_addr32[3];
		dst6->sin6_port = src6->sin6_port;
		return;
	default:
		break;
        }

	return;
}



/****
 * Scheduler:
 *
 * RR
 * LC
 * WRR
 * WLC
 */

/* Round robin scheduleur */
static int scheduler_rr(struct lbdr_server **server, int nb, int last)
{
	register uint32_t i;

	START(schd);

	i = last;

	DS(LOG(LOG_DEBUG, "(rr "));

	do {
		i++;
		i %= nb;
		if(server[i]->status == SRV_UP) {
			DS(LOG(LOG_DEBUG, "%i) ", i));
			STOP(schd);
			return i;
		}
	} while(i != last);

	DS(LOG(LOG_DEBUG, "-1) "));
	STOP(schd);
	return -1;
}

/* least connections scheduleur */
static int scheduler_lc(struct lbdr_server **s, int nb, int last)
{
	register int choice = -1, i;

	START(schd);

	i = last;

	DS(LOG(LOG_DEBUG, "(lc "));

	do {
		i++;
		i %= nb;
		if(s[i]->status == SRV_UP) {
			if(choice == -1) {
				choice = i;
			} else if(s[i]->active_cnx < s[choice]->active_cnx) {
				choice = i;
			}
		}
	} while(i != last);

	DS(LOG(LOG_DEBUG, "%i) ", choice));
	STOP(schd);
	return choice;
}


/* meme chose que RR mais prend en compte le poid */
static int scheduler_wrr(struct lbdr_server **server, int nb, int last)
{
	register int i;
	register int choice = -1;
	register uint32_t round = 1;
	register uint32_t total = 0;

	START(schd);
	DS(LOG(LOG_DEBUG, "(wrr ");)

	for(i = 0; i < nb; i++) {
		total += server[i]->weight;
		round += server[i]->data;
	}

	if(round > total) {
		round = 1;
	}

	for(i = 0; i < nb; i++) {
		if(round == 1) server[i]->data = 0;
		if(choice == -1) {
			/*if(((float)server[i]->data / (float)round) < ((float)server[i]->weight / (float)total)) {*/
			if(((float)server[i]->data * (float)total) < ((float)server[i]->weight * (float)round)) {
				choice = i;
				server[i]->data++;
			}
		}
	}

	DS(LOG(LOG_DEBUG, "%i)", choice);)
	STOP(schd);

	return choice;
}

/* Weighted Least Connections */
static int scheduler_wlc(struct lbdr_server **s, int nb, int last)
{
	register float tmp_load = 0.0;
	register float choice_load = FLT_MAX;
	register int choice = -1, i;

	i = last;

	START(schd);

	DS(LOG(LOG_DEBUG, "(wlc ");)

	do {
		i++;
		i %= nb;
		if(s[i]->status == SRV_UP && s[i]->weight > 0) {
			tmp_load = (s[i]->active_cnx / s[i]->weight);
			if(choice == -1) {
				choice = i;
				choice_load = tmp_load;
			} else if(tmp_load < choice_load) {
				choice = i;
				choice_load = tmp_load;
			}
		}
	} while(i != last);

	DS(LOG(LOG_DEBUG, "%i) ", choice));

	STOP(schd);
	return choice;
}



static int sched_number(int (*sched)(struct lbdr_server **, int, int))
{
	if(sched == scheduler_rr) {
		return LBDR_SCHED_RR;
	} else if(sched == scheduler_wrr) {
		return LBDR_SCHED_WRR;
	} else if(sched == scheduler_lc) {
		return LBDR_SCHED_LC;
	} else if(sched == scheduler_wlc) {
		return LBDR_SCHED_WLC;
	}

	return -1;
}



/***
 * manipulating service:
 *
 * register_service();
 * unregister_service();
 * find_service();
 * free_service();
 * clone_service();
 */


static struct lbdr_service *register_service(struct sockaddr_storage *ip, int sched, int tls, uint32_t htsize)
{
	struct lbdr_service *last = NULL;
	struct lbdr_service *tmp;
	int i;
	uint32_t h;

	DG(uprintf("Service %s... ", paddr(ip));)

	if(htsize > 25) htsize = 25;
	else if(htsize < 1) htsize = 1;

	htsize = 1 << htsize;

	h = sfhash(ip) & (HT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			DG(uprintf("Already alive.\n"));
			return tmp;
		}
		last = tmp;
		tmp = tmp->next;
	};

	DG(uprintf("No existent => ");)

	D(uprintf("New service %s(HConn: %i) ", tls ? "[SSL] " : "", htsize);)

	tmp = malloc(sizeof(struct lbdr_service), M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp) goto out;

	tmp->ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp->ip) goto free_tmp;

	tmp->htable = malloc(sizeof(struct lbdr_conn) * htsize, M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp->htable) goto free_ip;

	tmp->to = malloc(sizeof(struct counters), M_BUF, M_NOWAIT|M_ZERO);
	if(!tmp->to) goto free_htable;

	tmp->locktable = malloc(sizeof(kmutex_t) * htsize, M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp->locktable) goto free_to;


	for(i = 0; i < htsize; i++) {
		mtx_init(&(tmp->locktable[i]), "locktable", NULL, MTX_DEF);
		tmp->htable[i].status = 0;
	}

	copy_addr(tmp->ip, ip);
	tmp->htsize = htsize;
	tmp->server = NULL;
	tmp->next = NULL;
	tmp->clone = NULL;
	tmp->cloned = 0;
	tmp->tls = tls;
	tmp->nb_server = 0;
	tmp->to->nb_coll = 0;
	tmp->refcnt = 0;
	tmp->to->err_nosyn = 0;
	tmp->to->err_nosrv = 0;
	tmp->used_server = 0;
	tmp->to->t_syn = TIMEOUT_SYN * nticks;
	tmp->to->t_est = TIMEOUT_ESTABL * nticks;

	switch(sched) {
		case LBDR_SCHED_WLC:
			D(uprintf("Using Weighted-Least-Connections scheduler\n");)
			tmp->sched = scheduler_wlc;
			break;
		case LBDR_SCHED_WRR:
			D(uprintf("Using Weighted-Round-Robin scheduler\n");)
			tmp->sched = scheduler_wrr;
			break;
		case LBDR_SCHED_LC:
			D(uprintf("Using Least-Connections scheduler\n");)
			tmp->sched = scheduler_lc;
			break;
		default:
			D(uprintf("Unknow scheduler, ");)
		case LBDR_SCHED_RR:
			D(uprintf("Using Round-Robin scheduler\n");)
			tmp->sched = scheduler_rr;
			break;
	}

	if(!service_available[h]) {
		service_available[h] = tmp;
	} else /*if(last)*/ {
		last->next = tmp;
	}

	return tmp;

free_to:
	free(tmp->to, M_BUF);
free_htable:
	free(tmp->htable, M_BUF);
free_ip:
	free(tmp->ip, M_BUF);
free_tmp:
	free(tmp, M_BUF);
out:
	log(LOG_ERR, "drlb: not enough kernel memory!\n");
	return NULL;
}



/***
 * Fonction appele pour chaque paquet TCP.
 * I hope this is fast enough...
 */

static struct lbdr_service *find_service(struct sockaddr_storage *ip)
{
	struct lbdr_service *tmp;
	uint32_t h;

	START(fd_svi);

	h = sfhash(ip) & (HT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			STOP(fd_svi);
			return tmp;
		}
		tmp = tmp->next;
	};

	STOP(fd_svi);
	return NULL;
}



/*
 * Must lock the structure ?
 */
static void free_service(struct lbdr_service *tmp)
{
	struct lbdr_conn *tcnx;
	struct lbdr_conn *ncnx = NULL;
	int i;

	/* TODO: need to sleep() or callout this */
	while(tmp->refcnt > 0);

	if(tmp->cloned) {
		free(tmp->ip, M_BUF);
		free(tmp, M_BUF);
		return;
	}


	/* delete clone */
	FOREACHCLONE(tmp,
		while(clone->refcnt > 0);
		free(clone->ip, M_BUF);
		free(clone, M_BUF);
	)

	for(i = 0; i < tmp->htsize; i++) {
		tcnx = &(tmp->htable[i]);
		ncnx = tcnx->next;
		callout_drain(&(tcnx->cal));
		while(ncnx) {
			tcnx = ncnx;
			ncnx = tcnx->next;
			D(if(tcnx->status & (CNX_FLAG_ACT | CNX_FLAG_EXIST)) uprintf("+");
			  else if(tcnx->status & CNX_FLAG_EXIST) uprintf(".");)
			callout_drain(&(tcnx->cal));
			free(tcnx, M_BUF);
		}
		mtx_destroy(&(tmp->locktable[i]));
	}
	for(i = 0; i < tmp->nb_server; i++) {
		if(tmp->server[i]) {
			D(uprintf("*"));
			if(tmp->server[i]->rt)
				RTFREE(tmp->server[i]->rt);
			free(tmp->server[i]->ip, M_BUF);
			free(tmp->server[i], M_BUF);
		}
	}
	free(tmp->locktable, M_BUF);
	free(tmp->htable, M_BUF);
	free(tmp->server, M_BUF);
	free(tmp->to, M_BUF);
	free(tmp->ip, M_BUF);
	free(tmp, M_BUF);
}





static int unregister_service(struct sockaddr_storage *ip)
{
	struct lbdr_service *last = NULL;
	struct lbdr_service *tmp;
	uint32_t h;

	DG(uprintf("Service %s... ", paddr(ip));)

	h = sfhash(ip) & (HT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			break;
		}
		last = tmp;
		tmp = tmp->next;
	};

	if(!tmp) {
		DG(uprintf("Not found ?\n"));
		return -1;
	}

	D(uprintf("S");)

	if(!tmp->cloned) {
		FOREACHCLONE(tmp,
			unregister_service(clone->ip);
		)
	}

	if(service_available[h] == tmp) {
		service_available[h] = tmp->next;
	} else /*if(last)*/ {
		last->next = tmp->next;
	}

	DG(uprintf("Removed.\n"));

	return 0;
}



/*
 *
 */

static int clone_service(struct lbdr_service *service, struct sockaddr_storage *nip)
{
	struct lbdr_service *last = NULL;
	struct lbdr_service *tmp;
	struct lbdr_service *new;
	uint32_t h;

	DG(uprintf("Service %s... ", paddr(nip));)

	h = sfhash(nip) & (HT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, nip)) {
			DG(uprintf("Already alive.\n"));
			return 1;
		}
		last = tmp;
		tmp = tmp->next;
	};

	DG(uprintf("No existent => ");)
	D(uprintf("New service clone of %s\n", paddr(service->ip));)

	new = malloc(sizeof(struct lbdr_service), M_BUF, M_NOWAIT | M_ZERO);
	if(!new) return -1;

	memcpy(new, service, sizeof(struct lbdr_service));

	new->ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);
	if(!new->ip) return -1;

	copy_addr(new->ip, nip);
	new->cloned = 1;
	new->clone = NULL;
	new->next = NULL;

	if(!service_available[h]) {
		service_available[h] = new;
	} else /*if(last)*/ {
		last->next = new;
	}

	/* add to clone list */
	tmp = service->clone;

	if(!tmp) {
		service->clone = new;
	} else {
		while(tmp->clone) {
			tmp = tmp->clone;
		}
		tmp->clone = new;
	}

	return 0;
}



/***
 * manipulating servers:
 *
 * add_server();
 * delete_server();
 * modify_server();
 */


/* tableau de pointeurs sur liste serveurs, nb de serveur, nelle ip */
static struct lbdr_server ** add_server(struct lbdr_server **server, int nb, struct sockaddr_storage *s, int weight, int *err)
{
	struct lbdr_server **srv;
	int i = 0;

	if(nb >= MAX_SERVER) {
		D(uprintf("Can't add more than %i servers\n", MAX_SERVER);)
		*err = EOVERFLOW;
		return NULL;
	}

	DG(uprintf("Server %s... ", paddr(s));)

	while(i < nb) {
		if(server[i]) {
			if(same_addr(server[i]->ip, s)) {
				DG(uprintf("Already here.\n"));
				*err = EEXIST;
				return NULL;
			}
		}
		i++;
	}

	DG(uprintf("No existent => " );)
	D(uprintf("New server %s ", paddr(s));)

	srv = malloc(sizeof(struct lbdr_server *) * nb+1, M_BUF, M_NOWAIT | M_ZERO);
	if(!srv) goto out;

	for(i = 0; i < nb; i++) {
		srv[i] = server[i];
	}

	srv[nb] = malloc(sizeof(struct lbdr_server), M_BUF, M_NOWAIT | M_ZERO);
	if(!srv[nb]) goto free_srv;

	srv[nb]->ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);
	if(!srv[nb]->ip) goto free_srv_tab;

        srv[nb]->rt = NULL;
        srv[nb]->status = SRV_UP;
        srv[nb]->weight = weight;
        srv[nb]->inactive_cnx = 0;
        srv[nb]->active_cnx = 0;
        srv[nb]->nb_cnx = 0;
        srv[nb]->no_fin = 0;
        srv[nb]->pkts = 0;
        srv[nb]->bytes = 0;
        srv[nb]->err_unreach = 0;

	copy_addr(srv[nb]->ip, s);

	srv[nb]->rt = rtalloc1((struct sockaddr *)srv[nb]->ip, 1, 0UL);
#ifndef __NetBSD__
	if(srv[nb]->rt)
		RT_UNLOCK(srv[nb]->rt);
#endif

        if(!(srv[nb]->rt)) {
        	log(LOG_ERR, "drlb: Warning: Host %s is unreachable!\n", paddr(s));
        } else if(srv[nb]->rt->rt_flags & RTF_GATEWAY) {
        	log(LOG_ERR, "drlb: Warning: Host %s is not on my network!\n", paddr(s));
        }
        D(else uprintf("(route via %s)\n", srv[nb]->rt->rt_ifp->if_xname));

	*err = 0;
	return srv;

free_srv_tab:
	free(srv[nb], M_BUF);

free_srv:
	free(srv, M_BUF);

out:
	log(LOG_ERR, "drlb: not enough kernel memory!\n");
	*err = ENOMEM;
	return NULL;
}





static int modify_server(struct lbdr_server **server, struct sockaddr_storage *s, int nb, int weight, int status)
{
	int i = nb - 1;

	if(weight < 0 || weight > 100) {
		/* TODO not print when only modify status */
		D(uprintf("Invalid weight (%i). Weight must be between 0 and 100\n", weight);)
		return EINVAL;
	}

	D(uprintf("Server %s... ", paddr(s));)

	while(i >= 0) {
		if(!server[i]) {
			continue;
		}

		/* uh ? don't modify weight if inactivate only */
		if(same_addr(server[i]->ip, s)) {
			D(uprintf("Changed (w: %i  s: %i).\n", weight, status);)
			server[i]->status = status;
			server[i]->weight = weight;
			return 0;
		}
		i--;
	}

	D(uprintf("not found.\n");)

	return ENOENT;
}





static struct lbdr_server ** delete_server(struct lbdr_server **server, struct sockaddr_storage *s, int nb, int *err)
{
	struct lbdr_server **srv;
	int i = 0, dst = 0;

	nb--;

	D(uprintf("Rebuilding (%i) server table without %s... ", nb, paddr(s));)

	srv = malloc(sizeof(struct lbdr_server *) * nb, M_BUF, M_NOWAIT | M_ZERO);
	if(!srv) {
		log(LOG_ERR, "drlb: not enough kernel memory!\n");
		*err = ENOMEM;
		return NULL;
	}

	while(i <= nb) {
		if(same_addr(server[i]->ip, s)) {
			if(server[i]->active_cnx || server[i]->inactive_cnx) {
				D(uprintf("Can't delete a server with actives connections. Please deactivate it before.\n");)
				free(srv, M_BUF);
				*err = EBUSY;
				return NULL;
			}
			if(server[i]->rt)
				RTFREE(server[i]->rt);
			free(server[i]->ip, M_BUF);
			free(server[i], M_BUF);
			server[i] = NULL;
		} else {
			if(dst < nb)
				srv[dst] = server[i];
			dst++;
		}
		i++;
	}

	D(uprintf("%s\n", dst != i ? "OK" : "FAIL"));

	if(dst != i) {
		*err = 0;
		return srv;
	}

	free(srv, M_BUF);
	*err = ENOENT;
	return NULL;
}



/***
 * manipulating connections:
 * create_cnx();
 * find_cnx();
 * close_cnx();
 * drop_cnx();
 */

static struct lbdr_conn * create_cnx(struct lbdr_conn *ht, uint32_t h, struct lbdr_server *server, struct sockaddr_storage *ip, u_int *coll, kmutex_t *p)
{
	struct lbdr_conn *cnx;

	START(cr_cnx);

	cnx = &(ht[h]);

	if(cnx->status != 0) {
		while(cnx->next) cnx = cnx->next;
		cnx->next = malloc(sizeof(struct lbdr_conn), M_BUF, M_NOWAIT);
		cnx = cnx->next;
		cnx->next = NULL;
		(*coll)++;
		if(!cnx) {
			STOP(cr_cnx);
			return NULL;
		}
	}

	callout_init(&(cnx->cal), MPSAFE);
	cnx->status = CNX_SYN;
	cnx->srv = server;
	cnx->parent = ht;
	cnx->pmtx = p;
	copy_addr(&(cnx->ip_src), ip);
	cnx->hash = h;

	STOP(cr_cnx);
	return cnx;
}


static struct lbdr_conn *find_cnx(struct lbdr_conn *ht, uint32_t h, struct sockaddr_storage *ip)
{
	struct lbdr_conn *tmp;

	START(fd_cnx);

	tmp = &(ht[h]);

	while(tmp) {
		if(same_addr(&(tmp->ip_src), ip)) {
			STOP(fd_cnx);
			return tmp;
		}
		tmp = tmp->next;
	}
	
	STOP(fd_cnx);
	return NULL;
}


/* CALLOUT: on drop une cnx inactive */
static void drop_cnx(void *a)
{
	struct lbdr_conn *cnx = (struct lbdr_conn *)a;
	struct lbdr_conn *tmp;
	struct lbdr_conn *last = NULL;
	kmutex_t *mtx = cnx->pmtx;

	BIG_LOCK();
	mtx_lock(mtx);
	START(dr_cnx);

	if (callout_pending(&(cnx->cal))) {
		D(LOG(LOG_DEBUG, "drop_cnx(): pending\n");)
		goto unlock;
	}
	if (!callout_active(&(cnx->cal))) {
		D(LOG(LOG_DEBUG, "drop_cnx(): not active\n");)
		goto unlock;
	}
	callout_deactivate(&(cnx->cal));


	if(cnx->status & CNX_FLAG_ACT) {
		log(LOG_ERR, "drlb: drop_cnx() cnx already active!\n");
		cnx->srv->active_cnx--;
	} else {
		if(!(cnx->status & CNX_FLAG_DATA)) cnx->srv->no_fin++;
		cnx->srv->inactive_cnx--;
	}

	DT(LOG(LOG_DEBUG, "Timeout %s is dropped\n", paddr(&(cnx->ip_src))));

	tmp = &(cnx->parent[cnx->hash]);
	
	if(cnx == tmp) {
		cnx->srv = NULL;
		cnx->parent = NULL;
		cnx->hash = 0;
		cnx->status = 0;
		bzero(&(cnx->ip_src), sizeof(struct sockaddr_storage));
		goto unlock;
	} else {
		while(tmp->next) {
			last = tmp;
			tmp = tmp->next;
			if(tmp == cnx) {
				last->next = cnx->next;
				N(callout_destroy(&(cnx->cal));)
				free(cnx, M_BUF);
				goto unlock;
			}
		}
	}

	log(LOG_ERR, "drlb: drop_cnx() cnx not found !\n");
unlock:
	STOP(dr_cnx);
	mtx_unlock(mtx);
	BIG_UNLOCK();
}





/* CALLOUT: on passe la cnx en inactive */
static void close_cnx(void *a)
{
	struct lbdr_conn *cnx = (struct lbdr_conn *)a;

	BIG_LOCK();
	mtx_lock(cnx->pmtx);
	START(cl_cnx);

	if (callout_pending(&(cnx->cal))) {
		D(LOG(LOG_DEBUG, "close_cnx(): pending\n");)
		goto unlock;
	}
	if (!callout_active(&(cnx->cal))) {
		D(LOG(LOG_DEBUG, "close_cnx(): not active\n");)
		goto unlock;
	}
	callout_deactivate(&(cnx->cal));

	DT(LOG(LOG_DEBUG, "Timeout %s becomes INACTIVE (was %i)\n", paddr(&(cnx->ip_src)), cnx->status);)

	if(cnx->status & CNX_FLAG_ACT) {
		cnx->srv->active_cnx--;
		cnx->srv->inactive_cnx++;
	} else {
		log(LOG_ERR, "drlb: close_cnx() try to close inactive cnx!\n");
	}

	cnx->status = CNX_TIMEOUT;

	callout_reset(&(cnx->cal), t_clo, drop_cnx, (void *)cnx);

unlock:
	STOP(cl_cnx);
	mtx_unlock(cnx->pmtx);
	BIG_UNLOCK();
}



static __inline__ int new_cnx(struct lbdr_service *service, struct lbdr_conn **c, u_char thflags, struct sockaddr_storage *ip, uint32_t h)
{
	struct lbdr_conn *cnx;
	uint32_t srv;

	START(inp_new);

	if(!(thflags & TH_SYN)) {
		DT(LOG(LOG_DEBUG, "!S "));
		service->to->err_nosyn++;
		STOP(inp_new);
		return NOTHING;
	}
	DT(LOG(LOG_DEBUG, "S "));

	srv = (*service->sched)(service->server, service->nb_server, service->used_server);

	if(srv == -1) {
		DT(LOG(LOG_DEBUG, "!srv "));
		service->to->err_nosrv++;
		if(!((service->to->err_nosrv - 1) % LOGLIMIT)) log(LOG_ERR, "drlb: No server found!\n");
		STOP(inp_new);
		return NOTHING;
	}

	service->used_server = srv;

	cnx = create_cnx(service->htable, h, service->server[srv], ip, &(service->to->nb_coll), &(service->locktable[h]));

	if(!cnx) {
		info.err_nomem++;
		if(!((info.err_nomem - 1) % LOGLIMIT)) log(LOG_ERR, "drlb: not enough kernel memory!\n");
		STOP(inp_new);
		return DROP;
	}
	
	cnx->srv->nb_cnx++;
	cnx->srv->active_cnx++;
	callout_reset(&(cnx->cal), service->to->t_syn, close_cnx, (void *)cnx);
	*c = cnx;

	STOP(inp_new);
	return 0;
}


/**
 * change cnx status
 */
static __inline__ int update_status_cnx(struct lbdr_service *service, struct lbdr_conn *cnx, u_char thflags)
{
	uint32_t srv;

	START(inp_old);

	if(thflags & TH_SYN) {

		DT(LOG(LOG_DEBUG, "?S ");)

		if(cnx->status & CNX_FLAG_ACT) {
			cnx->srv->active_cnx--;
		} else {
			cnx->srv->inactive_cnx--;
		}

		srv = (*service->sched)(service->server, service->nb_server, service->used_server);

		if(srv == -1) {
			service->to->err_nosrv++;
		} else {
			service->used_server = srv;
			cnx->srv = service->server[srv];
		}

		cnx->srv->nb_cnx++;
		cnx->srv->active_cnx++;
		cnx->status = CNX_SYN;

		callout_reset(&(cnx->cal), service->to->t_syn, close_cnx, (void *)cnx);

	} else if(thflags & TH_FIN || thflags & TH_RST) {

		DT(LOG(LOG_DEBUG, "%c ", thflags & TH_FIN ? 'F' : 'R'));

		if(cnx->status & CNX_FLAG_ACT) { /* SYN or EST (active) */
			cnx->srv->active_cnx--;
			cnx->srv->inactive_cnx++;
		}

		cnx->status = CNX_CLOSE;

		callout_reset(&(cnx->cal), t_clo, drop_cnx, (void *)cnx);

	} else switch(cnx->status) {
		case CNX_TIMEOUT:
			cnx->srv->inactive_cnx--;
			cnx->srv->active_cnx++;
		case CNX_SYN:
			cnx->status = CNX_ESTABL;
		case CNX_ESTABL:
			callout_reset(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);
		default:
			break;
	}

	STOP(inp_old);

	return 0;
}


static __inline__ int update_status_ssl_cnx(struct lbdr_service *service, struct lbdr_conn *cnx, u_char thflags)
{
	START(inp_old);

	if(!(cnx->status & CNX_FLAG_ACT)) {
		cnx->srv->active_cnx++;
		cnx->srv->inactive_cnx--;
	}

	cnx->status = CNX_ESTABL;
	callout_reset(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);

	STOP(inp_old);

	return 0;
}



/**
 * Direct forwarding function
 * Do I need to check:
 *	- mbuf size (> struct ip[6]) ?
 *	- header lenght ?
 *	- checksum ? no!
 *	- packet lenght
 *	- traffic conditionner or firewall
 *	- fragment
 * I doesn't verify that pfil will do that for me..
 */
static __inline__ int forward(struct lbdr_conn *cnx, struct mbuf *mb)
{
	/*struct rtentry *rt;
	rt = cnx->srv->rt;*/
	START(inp_fwd);


	if (cnx->srv->rt) {
		if(cnx->srv->rt->rt_ifp && (cnx->srv->rt->rt_flags & RTF_UP))
			goto fwd;
		RTFREE(cnx->srv->rt);
		cnx->srv->rt = NULL;
	}
	cnx->srv->rt = rtalloc1((struct sockaddr *)cnx->srv->ip, 1, 0UL);

#ifndef __NetBSD__
	if(cnx->srv->rt)
		RT_UNLOCK(cnx->srv->rt);
#endif

fwd:
	if(!cnx->srv->rt ||
		(cnx->srv->rt->rt_flags & RTF_GATEWAY)
		/*|| !(cnx->srv->rt->rt_ifp->if_flags & IFF_UP) Is that util ? */
		) {
		DT(LOG(LOG_DEBUG, "Unreachable ");)
		/* icmp or redirect to backup_server
		icmp_error(m, ICMP_UNREACH, ICMP_UNREACH_HOST, 0, 0); */
		cnx->srv->err_unreach++;

		if(!((cnx->srv->err_unreach - 1) % LOGLIMIT))
			log(LOG_ERR, "drlb: Host %s unreachable !\n", paddr(cnx->srv->ip));

		STOP(inp_fwd);
		return NOTHING;
	}

	cnx->srv->rt->rt_rmx.rmx_pksent++;

	DT(LOG(LOG_DEBUG, "=> %s (%s) ", paddr(cnx->srv->ip), cnx->srv->rt->rt_ifp->if_xname);)

	(*cnx->srv->rt->rt_ifp->if_output)(cnx->srv->rt->rt_ifp, mb, (struct sockaddr *)cnx->srv->ip, cnx->srv->rt);

	STOP(inp_fwd);
	return 0;
}


/***
 * main functions (pfil entry)
 */


/**
 * PFIL input for IPv4
 */
static int
F(lb4_input(void *arg, struct mbuf **m, struct ifnet *i, int dir, struct inpcb *inp))
N(lb4_input(void *arg, struct mbuf **m, struct ifnet *i, int dir))
{
        struct mbuf *mb;
        struct ip *iph;
	struct sockaddr_storage ss;
	struct sockaddr_in *in;
	struct lbdr_service *service;
	struct lbdr_conn *cnx;
	kmutex_t *mtx;
	uint32_t hash;
	union protohdr header;

	mb = *m;
	iph = mtod(mb, struct ip *);

        if (iph->ip_p != IPPROTO_TCP) return 0;

	if (mb->m_len < sizeof (struct ip) + sizeof(struct tcphdr)) {
		if ((mb = m_pullup(mb, sizeof (struct ip) + sizeof(struct tcphdr))) == NULL) {
			DG(LOG(LOG_DEBUG, "mbuf too short");)
			info.err_badmbuf++;
			return 0;
		}
	}

        header.raw = (char *)iph + (iph->ip_hl << 2);

	in = (struct sockaddr_in *)&ss;
	in->sin_len = sizeof(struct sockaddr_in);
	in->sin_family = AF_INET;
	in->sin_port = header.portp[1];
	in->sin_addr = iph->ip_dst;

	service = find_service((struct sockaddr_storage *)in);

	if(!service) return 0;

	BIG_LOCK();
	service->refcnt++;

        DT({
		in->sin_port = header.portp[0];
		in->sin_addr = iph->ip_src;
        	LOG(LOG_DEBUG, "%s %s/%s -> ", i->if_xname, proto_name(iph->ip_p), paddr((struct sockaddr_storage *)in));
		in->sin_port = header.portp[1];
		in->sin_addr = iph->ip_dst;
        	LOG(LOG_DEBUG, "%s ", paddr((struct sockaddr_storage *)in));
	})


	in->sin_addr = iph->ip_src;
	in->sin_port = service->tls ? 0 : header.portp[0];

	hash = sfhash((struct sockaddr_storage *)in) & (service->htsize-1);
	mtx = &(service->locktable[hash]);
	mtx_lock(mtx);

	cnx = find_cnx(service->htable, hash, (struct sockaddr_storage *)in);

	if(!cnx) {
		switch(new_cnx(service, &cnx, header.th->th_flags, (struct sockaddr_storage *)in, hash)) {
			case DROP:
				goto drop;
			case NOTHING:
				goto nothing;
			default:
				break;
		}
		
	} else { 
		if(service->tls)
			update_status_ssl_cnx(service, cnx, (header.th)->th_flags);
		else
			update_status_cnx(service, cnx, (header.th)->th_flags);
	}

	iph->ip_len = htons(iph->ip_len);
	iph->ip_off = htons(iph->ip_off);

	if(forward(cnx, mb) == NOTHING) goto nothing;

        cnx->srv->pkts++;
        cnx->srv->bytes += mb->m_len;

        *m = NULL;

	DT(LOG(LOG_DEBUG, "\n");)

	/*mtx_unlock(&(service->locktable[hash]));*/
	mtx_unlock(mtx);
	service->refcnt--;
	BIG_UNLOCK();

	return 0;

drop:
	DT(LOG(LOG_DEBUG, "DROP ");)
	m_freem(*m);
        *m = NULL;

nothing:
	DT(LOG(LOG_DEBUG, "PASS\n");)
	service->refcnt--;
	mtx_unlock(&(service->locktable[hash]));

        BIG_UNLOCK();
        return 0;
}


/**
 * PFIL entry for IPv6
 */
static int
F(lb6_input(void *arg, struct mbuf **m, struct ifnet *i, int dir, struct inpcb *inp))
N(lb6_input(void *arg, struct mbuf **m, struct ifnet *i, int dir))
{
        struct mbuf *mb;
        struct ip6_hdr *iph6;
	struct sockaddr_storage ss;
	struct sockaddr_in6 *in6;
	struct lbdr_service *service;
	struct lbdr_conn *cnx;
	uint32_t hash;
	union protohdr header;

	mb = *m;
	iph6 = mtod(mb, struct ip6_hdr *);

        if (iph6->ip6_nxt != IPPROTO_TCP) return 0;

	if (mb->m_len < sizeof (struct ip6_hdr) + sizeof(struct tcphdr)) {
		if ((mb = m_pullup(mb, sizeof (struct ip6_hdr) + sizeof(struct tcphdr))) == NULL) {
			DG(LOG(LOG_DEBUG, "mbuf too short");)
			info.err_badmbuf++;
			return 0;
		}
	}

        header.raw = (char *)iph6 + sizeof(struct ip6_hdr);

	in6 = (struct sockaddr_in6 *)&ss;
	in6->sin6_len = sizeof(struct sockaddr_in6);
	in6->sin6_family = AF_INET6;
	in6->sin6_port = header.portp[1];
	in6->sin6_addr = iph6->ip6_dst;

	service = find_service((struct sockaddr_storage *)in6);

	if(!service) return 0;

	BIG_LOCK();
	service->refcnt++;

        DT({
		in6->sin6_port = header.portp[0];
		in6->sin6_addr = iph6->ip6_src;
        	LOG(LOG_DEBUG, "%s %s/%s -> ", i->if_xname, proto_name(iph6->ip6_nxt), paddr((struct sockaddr_storage *)in6));
		in6->sin6_port = header.portp[1];
		in6->sin6_addr = iph6->ip6_dst;
        	LOG(LOG_DEBUG, "%s ", paddr((struct sockaddr_storage *)in6));
	})

	in6->sin6_addr = iph6->ip6_src;
	in6->sin6_port = service->tls ? 0 : header.portp[0];

	hash = sfhash((struct sockaddr_storage *)in6) & (service->htsize-1);
	mtx_lock(&(service->locktable[hash]));

	cnx = find_cnx(service->htable, hash, (struct sockaddr_storage *)in6);

	if(!cnx) {

		switch(new_cnx(service, &cnx, header.th->th_flags, (struct sockaddr_storage *)in6, hash)) {
			case DROP:
				goto drop;
			case NOTHING:
				goto nothing;
			default:
				break;
		}
		
	} else { 
		if(service->tls)
			update_status_ssl_cnx(service, cnx, (header.th)->th_flags);
		else
			update_status_cnx(service, cnx, (header.th)->th_flags);
	}

	/* need to modify mbuf ? */

	if(forward(cnx, mb) == NOTHING) goto nothing;

        cnx->srv->pkts++;
        cnx->srv->bytes += mb->m_len;

        *m = NULL;

	DT(LOG(LOG_DEBUG, "\n");)

	mtx_unlock(&(service->locktable[hash]));
	service->refcnt--;
	BIG_UNLOCK();

	return 0;

drop:
	DT(LOG(LOG_DEBUG, "DROP ");)
	m_freem(*m);
        *m = NULL;

nothing:
	DT(LOG(LOG_DEBUG, "PASS\n");)
	service->refcnt--;

	mtx_unlock(&(service->locktable[hash]));
        BIG_UNLOCK();

        return 0;
}


/***
 * config and statistique:
 *
 * dump_stats();
 * ioctl();
 */
static u_int get_nb_server(uint32_t *nb_cnx)
{
	struct lbdr_service *s = NULL;
	struct lbdr_server **srv;
	int i, r = 0, j = 0;;

	for(i = 0; i < HT_SERVICE; i++) {
		s = service_available[i];
		while(s) {
			if(s->cloned) {
				r++;
			} else {
				r += s->nb_server;
				srv = s->server;
				for(j = 0; j < s->nb_server; j++) {
					*nb_cnx += srv[j]->nb_cnx;
				}
			}
			s = s->next;
		}
	}

	return r;
}


static int dump_stats(struct lbdr_stats *stats)
{
	struct lbdr_service *s = NULL, *clone;
	struct lbdr_server **srv = NULL;
	struct lbdr_param *dest, prm;
	int i, j, max = 0, nb = 0;

	max = stats->size / sizeof(struct lbdr_param);

	dest = stats->data;

	for(i = 0; i < HT_SERVICE; i++) {
		s = service_available[i];
		while(s) {
			if(!s->cloned) {
			clone = s->clone;
			while(clone) {
				bzero(&prm, sizeof(struct lbdr_param));
				copy_addr(&(prm.ip_dst), s->ip);
				copy_addr(&(prm.ip_srv), clone->ip);
				prm.clone = 1;
				prm.sched = sched_number(s->sched);
				prm.coll = s->to->nb_coll;
				prm.tls = s->tls;
				prm.t_syn = s->to->t_syn / nticks;
				prm.err_nosyn = s->to->err_nosyn;
				prm.err_nosrv = s->to->err_nosrv;
				prm.t_est = s->to->t_est / nticks;
				prm.t_clo = t_clo / nticks;
				prm.hashtablesize = s->htsize;
				if(nb < max) {
					if(copyout(&prm, &(dest[nb]), sizeof(struct lbdr_param)))
						return EFAULT;
					nb++;
				} else {
					D(uprintf("Not enough memory allocated. Can't dump all data!\n");)
					return EIO;
				}
				clone = clone->clone;
			}
			srv = s->server;
			for(j = 0; j < s->nb_server; j++) {
				/*in = (struct sockaddr_storage *)(srv[j]->ip);*/
				bzero(&prm, sizeof(struct lbdr_param));
				copy_addr(&(prm.ip_dst), s->ip);
				copy_addr(&(prm.ip_srv), srv[j]->ip);
				prm.weight = srv[j]->weight;
				prm.act_cnx = srv[j]->active_cnx;
				prm.inact_cnx = srv[j]->inactive_cnx;
				prm.status = srv[j]->status;
				prm.pkts = srv[j]->pkts;
				prm.bytes = srv[j]->bytes;
				prm.no_fin = srv[j]->no_fin;
				prm.err_unreach = srv[j]->err_unreach;
				prm.sched = sched_number(s->sched);
				prm.tot_cnx = srv[j]->nb_cnx;
				prm.coll = s->to->nb_coll;
				prm.tls = s->tls;
				prm.t_syn = s->to->t_syn / nticks;
				prm.err_nosyn = s->to->err_nosyn;
				prm.err_nosrv = s->to->err_nosrv;
				prm.t_est = s->to->t_est / nticks;
				prm.t_clo = t_clo / nticks;
				prm.hashtablesize = s->htsize;
				if(nb < max) {
					if(copyout(&prm, &(dest[nb]), sizeof(struct lbdr_param)))
						return EFAULT;
					nb++;
				} else {
					D(uprintf("Not enough memory allocated. Can't dump all data!\n");)
					return EIO;
				}
			}
			}
			s = s->next;
		}
	}

	return 0;
}

D(static void dump_time(void)
{
	uprintf("Average time used:\n"
		  "   %9llu msec in find_service() (%9u calls) %8u nsec/call\n"
		  "   %9llu msec in find_cnx()     (%9u calls) %8u nsec/call\n"
		  "   %9llu msec for existing cnx  (%9u calls) %8u nsec/call\n"
		  "   %9llu msec for new cnx       (%9u calls) %8u nsec/call\n"
		  "   %9llu msec in create_cnx()   (%9u calls) %8u nsec/call\n"
		  "   %9llu msec in scheduler()    (%9u calls) %8u nsec/call\n"
		  "   %9llu msec for forward       (%9u calls) %8u nsec/call\n"
		  "   %9llu msec in close_cnx()    (%9u calls) %8u nsec/call\n"
		  "   %9llu msec in drop_cnx()     (%9u calls) %8u nsec/call\n" ,
		fd_svi.ns/1000000, fd_svi.nb, fd_svi.nb > 0 ? (uint32_t)((float)fd_svi.ns / (float)fd_svi.nb) : 0,
		fd_cnx.ns/1000000, fd_cnx.nb, fd_cnx.nb > 0 ? (uint32_t)((float)fd_cnx.ns / (float)fd_cnx.nb) : 0,
		inp_old.ns/1000000, inp_old.nb, inp_old.nb > 0 ? (uint32_t)((float)inp_old.ns / (float)inp_old.nb) : 0,
		inp_new.ns/1000000, inp_new.nb, inp_new.nb > 0 ? (uint32_t)((float)inp_new.ns / (float)inp_new.nb) : 0,
		cr_cnx.ns/1000000, cr_cnx.nb, cr_cnx.nb > 0 ? (uint32_t)((float)cr_cnx.ns / (float)cr_cnx.nb) : 0,
		schd.ns/1000000, schd.nb, schd.nb > 0 ? (uint32_t)((float)schd.ns / (float)schd.nb) : 0,
		inp_fwd.ns/1000000, inp_fwd.nb, inp_fwd.nb > 0 ? (uint32_t)((float)inp_fwd.ns / (float)inp_fwd.nb) : 0,
		cl_cnx.ns/1000000, cl_cnx.nb, cl_cnx.nb > 0 ? (uint32_t)((float)cl_cnx.ns / (float)cl_cnx.nb) : 0,
		dr_cnx.ns/1000000, dr_cnx.nb, dr_cnx.nb > 0 ? (uint32_t)((float)dr_cnx.ns / (float)dr_cnx.nb) : 0 );
})

/* 
 * Device functions
 */


N(static int drlb_ioctl(dev_t d, u_long cmd, void *arg, int mode, struct lwp *lwp))
F(int drlb_ioctl(struct cdev *dev, u_long cmd, caddr_t arg, int mode, struct thread *td))
{
	int error = 0;
	struct lbdr_service * service, *last;
	struct lbdr_server **newsrv, **oldsrv;
	struct lbdr_param *prm = (struct lbdr_param *)arg;
	struct lbdr_timeout *to = (struct lbdr_timeout *)arg;
	struct lbdr_stats *stats = (struct lbdr_stats *)arg;
	struct lbdr_info *inf = (struct lbdr_info *)arg;
	struct sockaddr_storage *dst = &(prm->ip_dst);
	int i;

        switch(cmd) {
        case LBDRADDSRV:
		DG(LOG(LOG_DEBUG, "LBDRADDSRV\n"));
                service = register_service(dst, prm->sched, prm->tls, prm->hashtablesize);
                if(service) {
			if(service->cloned) {
				D(uprintf("Cannot change cloned service\n");)
				error = EPERM;
				break;
			}
			oldsrv = service->server;
			/* TODO: ip_srv->port = ip_dst->port */
			newsrv = add_server(oldsrv, service->nb_server, &(prm->ip_srv), prm->weight, &error);
			if(newsrv) {
				service->server = newsrv;
				service->nb_server++;
				FOREACHCLONE(service,
					clone->server = service->server;
					clone->nb_server = service->nb_server
				)
				free(oldsrv, M_BUF);
			}
		} else error = ENOMEM; /* only case of failure for register_service() */
                break;
        case LBDRDELSRV:
		DG(LOG(LOG_DEBUG, "LBDRDELSRV\n"));
		service = find_service(dst);
                if(service) {
			if(service->cloned) {
				D(uprintf("Cannot change cloned service\n");)
				error = EPERM;
				break;
			}
			oldsrv = service->server;
			newsrv = delete_server(oldsrv, &(prm->ip_srv), service->nb_server, &error);
			if(newsrv) {
				if(service->nb_server == 1) {
					unregister_service(dst);
					free(newsrv, M_BUF);
					free_service(service);
				} else {
					service->nb_server--;
					service->used_server %= service->nb_server;
					service->server = newsrv;
					FOREACHCLONE(service,
						clone->nb_server = service->nb_server;
						clone->used_server %= clone->nb_server;
						clone->server = service->server;
					)
					free(oldsrv, M_BUF);
				}
			}
                } else error = ENOENT;
                break;
        case LBDRCHGSVICE:
        	DG(LOG(LOG_DEBUG, "LBDRCHGSVICE\n");)
        	service = find_service(dst);
        	if(service) {
			if(service->cloned) {
				D(uprintf("Cannot change cloned service\n");)
				error = EPERM;
				break;
			}

			D(if(service->tls != prm->tls)
				uprintf("Switching to %sSSL\n", prm->tls ? "" : "no-");)
			service->tls = prm->tls;

			if(prm->sched == -1) 
				break;

			switch(prm->sched) {
			case LBDR_SCHED_WLC:
				D(uprintf("Switching to Weighted-Least-Connections scheduler\n");)
				service->sched = scheduler_wlc;
				break;
			case LBDR_SCHED_WRR:
				D(uprintf("Swtching to Weighted-Round-Robin scheduler\n");)
				service->sched = scheduler_wrr;
				break;
			case LBDR_SCHED_LC:
				D(uprintf("Swtching to Least-Connections scheduler\n");)
				service->sched = scheduler_lc;
				break;
			case LBDR_SCHED_RR:
				D(uprintf("Switching to Round-Robin scheduler\n");)
				service->sched = scheduler_rr;
				break;
			default:
				error = EINVAL;
				break;
			}
			FOREACHCLONE(service,
				clone->sched = service->sched;
				clone->tls = service->tls;
			)
        	} else error = ENOENT;
        	break;
        case LBDRCHGSRV:
		DG(LOG(LOG_DEBUG, "LBDRCHGSRV\n");)
		service = find_service(dst);
        	if(service) {
        		error = modify_server(service->server, &(prm->ip_srv), service->nb_server, prm->weight, prm->status);
        	} else error = ENOENT;
        	break;
        case LBDRCLONE:
        	DG(LOG(LOG_DEBUG, "LBDRCLONE\n");)
		service = find_service(dst);
        	if(service) {
			if(service->cloned) {
				D(uprintf("Cannot clone already cloned service\n");)
				error = EPERM;
				break;
			}
        		clone_service(service, &(prm->ip_srv));
        	} else error = ENOENT;
        	break;
        case LBDRINFO:
		inf->nb_cnx = 0;
		inf->nb_srv = get_nb_server(&(inf->nb_cnx));
		inf->err_nomem = info.err_nomem;
		inf->err_badmbuf = info.err_badmbuf;
		D(dump_time();)
        	break;
        case LBDRSTATS:
                error = dump_stats(stats);
		break;
	case LBDRCHGTO:
		DG(LOG(LOG_DEBUG, "LBDRCHGTO\n");)
		dst = &(to->ip_dst);
		service = find_service(dst);
		if(service) {
			if(service->cloned) {
				D(uprintf("Cannot change cloned service\n");)
				error = EPERM;
				break;
			}
			service->to->t_est = to->nt_est * nticks;
			service->to->t_syn = to->nt_syn * nticks;
			if(to->nt_clo != 0)
				t_clo = to->nt_clo * nticks;
			D(uprintf("Changing timeout Syn: %i\tEstablish: %i\tClosed: %i\n", service->to->t_syn/nticks, service->to->t_est/nticks, t_clo/nticks);)
		} else error = ENOENT;
		break;
	case LBDRCLEAR:
		DG(LOG(LOG_DEBUG, "LBDRCHGCLEAR\n");)
		info.nb_srv = 0;
		info.nb_cnx = 0;
		info.err_nomem = 0;
		info.err_badmbuf = 0;
		INIT(inp_fwd, "timefwd");
		INIT(inp_new, "timenew");
		INIT(inp_old, "timeold");
		INIT(fd_svi, "timefdsvi");
		INIT(fd_cnx, "timefdcnx");
		INIT(cl_cnx, "timeclcnx");
		INIT(dr_cnx, "timedrcnx");
		INIT(cr_cnx, "timecrcnx");
		INIT(schd, "timeschd");
		i = 0;
		while(i < HT_SERVICE) {
			service = service_available[i];
			while(service) {
				last = service;
				service = service->next;
				if(!last->cloned) {
					unregister_service(last->ip);
					free_service(last);
				}
			}
			i++;
		}
		bzero(service_available, sizeof(struct lbdr_service) * HT_SERVICE);
                break;
        default:
                D(LOG(LOG_DEBUG, "Unknow ioctl received\n");)
                error = EINVAL;
		break;
	}

        return (error);
}

/***
 * Module functions
 *
 * (load, unload, init, deinit)
 *
 */

static int init_module(void)
{
	F(struct timeval tv;)

        if (kld_loaded) return EEXIST;

        N(malloc_type_attach(M_BUF);)

	G(mtx_init(&big_lock, "mtxbig", NULL, MTX_DEF);)
	D(mtx_init(&mtx_log, "mtxlog", NULL, MTX_DEF);)

        F(sdev = make_dev(&drlb_devsw, 0, UID_ROOT, GID_WHEEL, 0600, "lbdr");)

	service_available = malloc(sizeof(struct lbdr_service) * HT_SERVICE, M_BUF, M_NOWAIT | M_ZERO);

	if(!service_available) { 
		log(LOG_ERR, "drlb: not enough kernel memory!\n");
		return ENOMEM;
	}

	N(nticks = mstohz(1000);)
	F(tv.tv_sec = 1; tv.tv_usec = 0; nticks = tvtohz(&tv);)

	kld_loaded = 1;
	log(LOG_INFO, "drlb: inserted (pfil)" D(" With Debug") DG(", Verbose") DT(", Trace") "\n");
	uprintf("drlb kld version %s. %iHz. HSservice: %i" D(" With Debug") DG(", Verbose") DT(", Trace") "\n", LBDR_VERSION, nticks, HT_SERVICE);
#ifndef NOBIG
	log(LOG_INFO, "drlb: [BIG-LOCKED]\n");
#endif

	t_clo = TIMEOUT_CLOSE * nticks;

	info.nb_srv = 0;
	info.nb_cnx = 0;
	info.err_nomem = 0;
	info.err_badmbuf = 0;

	INIT(inp_fwd, "timefwd");
	INIT(inp_new, "timenew");
	INIT(inp_old, "timeold");
	INIT(fd_svi, "timefdsvi");
	INIT(fd_cnx, "timefdcnx");
	INIT(cl_cnx, "timeclcnx");
	INIT(dr_cnx, "timedrcnx");
	INIT(cr_cnx, "timecrcnx");
	INIT(schd, "timeschd");

        return 0;
}

static int deinit_module(void)
{
	struct lbdr_service *tmp, *last;
	int i = HT_SERVICE - 1;

        if(!kld_loaded) return 0;

	while(i) {
		tmp = service_available[i];
		while(tmp) {
			last = tmp;
			tmp = tmp->next;
			if(!last->cloned) {
				unregister_service(last->ip);
				free_service(last);
			}
		}
		i--;
	}

        log(LOG_INFO, "drlb: unloaded\n");

	F(destroy_dev(sdev));
	D(mtx_destroy(&mtx_log);)
	G(mtx_destroy(&big_lock);)

	DEINIT(inp_fwd);
	DEINIT(inp_new);
	DEINIT(inp_old);
	DEINIT(fd_svi);
	DEINIT(fd_cnx);
	DEINIT(cl_cnx);
	DEINIT(dr_cnx);
	DEINIT(cr_cnx);
	DEINIT(schd);

	free(service_available, M_BUF);
        N(malloc_type_detach(M_BUF);)

        kld_loaded = 0;

        return 0;
}



N(static int lb_load(struct lkm_table *lkmtp, int what))
F(static int lb_load(struct module *m, int what, void *arg))
{
	struct pfil_head *pfh_inet4, *pfh_inet6;
	int err = 0;

	pfh_inet4 = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	pfh_inet6 = pfil_head_get(PFIL_TYPE_AF, AF_INET6);
	if (!pfh_inet4 || !pfh_inet6) return ESRCH;

	switch(what) {
	case N(LKM_E_LOAD) F(MOD_LOAD) :

		err = init_module();
		if(err) return err;
		pfil_add_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
		pfil_add_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);

		break;


	case N(LKM_E_UNLOAD) F(MOD_UNLOAD) :

		pfil_remove_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
		pfil_remove_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);
		err = deinit_module(); /* TODO: and what if error ? */

		break;

#ifdef __NetBSD__
	case LKM_E_STAT:
		break;
#else
	case MOD_QUIESCE:
		break;  /* TODO: return EBUSY if in use */
	case MOD_SHUTDOWN :
		pfil_remove_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
		pfil_remove_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);
		break;
#endif

	default:
		err = EINVAL;
		break;
	}

	return err;
}


#ifdef __NetBSD__


static int drlb_open(dev_t device, int flags, int fmt, struct lwp *process)
{
	return 0;
}

static int drlb_close(dev_t device, int flags, int fmt, struct lwp *process)
{
	return 0;
}

static struct cdevsw drlb_devsw = {
	drlb_open,
	drlb_close,
	(dev_type_read((*))) enodev,
	(dev_type_write((*))) enodev,
	drlb_ioctl,
	(dev_type_stop((*))) enodev, 0,
	(dev_type_poll((*))) enodev,
	(dev_type_mmap((*))) enodev, 0
};

/* dynamic majors numbers. */
MOD_DEV("drlb", "drlb", NULL, -1, &drlb_devsw, -1)

int drlb_lkmentry(struct lkm_table *lkmtp, int cmd, int ver)
{
        DISPATCH(lkmtp, cmd, ver, lb_load, lb_load, lb_load);
}


#else /* FreeBSD */

static moduledata_t drlb_mod = { "drlb", lb_load, NULL };
DECLARE_MODULE(drlb, drlb_mod, SI_SUB_KLD, SI_ORDER_ANY);
MODULE_VERSION(drlb, 1);

#endif


/* EOF */
