/*-
 * Copyright (c) 2008 Jean-Yves Moulin <jym@ldh.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/event.h>
#ifdef __NetBSD__
#include <util.h>
#else
#include <sys/param.h>
#include <libutil.h>
#endif
#include <pwd.h>
#include <grp.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef USE_CURL
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>
#endif


#include "lbdr.h"

#define MAXNAMESIZE	 255
#define MAXBUFSIZE	2048

/* All timeout are in seconds */

#define CONFIG_SERVICE 0x01
#define CONFIG_SERVER  0x02

#define DAEMON_SCHED_WRRL 0x10
#define DAEMON_SCHED_WLCL 0x20
#define DAEMON_SCHED_RD	  0x40

#ifdef USE_RRDTOOLS
#include <rrd.h>
#endif

#define ADDR(addr) ((unsigned char *)&addr)[0], ((unsigned char *)&addr)[1], ((unsigned char *)&addr)[2], ((unsigned char *)&addr)[3]

#ifndef __NetBSD__
struct pidfh *ph;
#endif

struct data {
	char *memory;
	size_t size;
};


struct lb_service {
	struct sockaddr_in *ip;
	struct sockaddr_in *ip_srv;
	struct lb_service *next;
	char *rrdname;
	char *rrdfile;
	int hashsize;
	int weight;
	int sched;
	int cnx;
	int tls;
	int status;
	int cloned;
	int syn_to;
	int est_to;
	unsigned short srv_port;
};

struct lb_server {
	struct sockaddr_in *ip;
	struct lb_server *next;
	struct lb_service *lbdr_service;
	pthread_t pid;
	int status;
	int load;
	int error;
	uint32_t conn_to;
	uint32_t recv_to;
	uint32_t retry_to;
	uint32_t nb_retry;
	uint32_t wait_to;
	uint32_t total_errors;
	uint32_t downtime;
	uint32_t total_downtime;
	char *file;
};

extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;
extern int optreset;
extern int errno;

static struct lb_server *server_available = NULL;
static int debug = 0;
static int use_clone = 0;
static int process = 1;
static int reread = 0;
static int use_syslog = 1;
static int loglevel = 0;
static int lb_dev = 0;
static char *user = NULL, *group = NULL, *mailserver = NULL, *email = NULL, *pid_file = NULL, *useragent = NULL, *hostname = NULL;
static pthread_mutex_t *mtx_ioctl;
static struct lbdr_info info;
static struct lbdr_stats stats;

/* Default timeout */
static int def_conn_to = 1;
static int def_recv_to = 2;
static int def_retry_to = 3;
static int def_nb_retry = 3;
static int def_wait_to = 10;

/* Default timeout for drlb: */
static int def_syn_to = 5;
static int def_est_to = 120;
static int def_clo_to = 60;




static int is_not_ip_valid(char *str)
{
	int i;

	for(i = 0; str[i]; i++) {
		if(str[i] != '0' && str[i] != '1' && str[i] != '2' && str[i] != '3' && str[i] != '4' && str[i] != '5' && str[i] != '6' && str[i] != '7' && str[i] != '8' && str[i] != '9' && str[i] != '.') return -1;
	}

	return 0;
}



static char * read_word(FILE *file)
{
	char c;
	char *buf = malloc(MAXNAMESIZE);
	char *ptr = buf;
	int comment = 0;

	bzero(buf, MAXNAMESIZE);

	while(fread(&c, 1, 1, file) && (ptr - buf) < MAXNAMESIZE) {
		if(comment && c != '\n') {
			continue;
		} else if(comment && c == '\n') {
			comment = 0;
			continue;
		} else if(c == '#') {
			comment = 1;
			continue;
		} else if(c == ' ' || c == '\t' || c == '\n') {
			if(ptr - buf == 0) continue;
			*ptr = 0;
			return buf;
		}
		*ptr = c;
		ptr++;
	}

	free(buf);
	return NULL;
}
static char * read_spaceword(FILE *file)
{
	char c;
	char *buf = malloc(MAXNAMESIZE);
	char *ptr = buf;
	int comment = 0;

	bzero(buf, MAXNAMESIZE);

	while(fread(&c, 1, 1, file) && (ptr - buf) < MAXNAMESIZE) {
		if(comment && c != '\n') {
			continue;
		} else if(comment && c == '\n') {
			comment = 0;
			continue;
		} else if(c == '#') {
			comment = 1;
			continue;
		} else if(c == ' ' || c == '\t' || c == '\n') {
			if(ptr - buf == 0) continue;
			*ptr++ = ' ';
			*ptr = 0;
			return buf;
		}
		*ptr = c;
		ptr++;
	}

	free(buf);
	return NULL;
}

static void printlog(int level, char *s, ...)
{
	char buf[MAXBUFSIZE];
	va_list ap;

	if(level > loglevel) return;

	va_start(ap, s);
	vsnprintf(buf, MAXBUFSIZE, s, ap);
	va_end(ap);

	if(use_syslog)
		syslog(LOG_INFO, buf);
	else
		puts(buf);
}





static int create_pid(char *p)
{
#ifdef __NetBSD__
	if(pidfile(p)) {
		printlog(0, "Can't create pidfile: %s. Exiting..\n", strerror(errno));
		return -1;
	}
#else /* FREEBSD */
	pid_t opi;
	ph = pidfile_open(p, 0600, &opi);
	if(ph == NULL) {
		if(errno == EEXIST) {
			printlog(0, "Daemon already running, pid %d. Exiting..", opi);
		} else {
			printlog(0, "Can't create pidfile: %s. Exiting..\n", strerror(errno));
		}
		return -1;
	}
	pidfile_write(ph);
#endif
	return 0;
}

static void remove_pid(void)
{
#ifndef __NetBSD__
	pidfile_remove(ph);
#endif
}




static int lb_ioctl(struct sockaddr_in *service, struct sockaddr_in *server, int type, int status, int sched, int weight, int tls, int h)
{
	struct lbdr_param *prm;

	prm = malloc(sizeof(struct lbdr_param));
	bzero(prm, sizeof(struct lbdr_param));

	memcpy(&(prm->ip_dst), service, sizeof(struct sockaddr_in));
	memcpy(&(prm->ip_srv), server, sizeof(struct sockaddr_in));

	prm->weight = weight;
	prm->sched = sched;
	prm->status = status;
	prm->tls = tls;
	prm->hashtablesize = h;

	pthread_mutex_lock(mtx_ioctl);

	if(ioctl(lb_dev, type, prm) == -1) {
		printlog(0, "ioctl: %s", strerror(errno));
	}

	pthread_mutex_unlock(mtx_ioctl);

	free(prm);

	return 0;
}


/* get lbdr server list 
static int update_status(void)
{
	struct lbdr_param *prm;
        static struct lbdr_stats stats;
	int nb_srv = 0;
	int i;

	i = ioctl(lb_dev, LBDRGETSRV, &nb_srv);
	if(i == -1) {
		printlog(0, "ioctl error: %s", strerror(errno));
	}

	stats.s = nb_srv * sizeof(struct lbdr_param);
	stats.data = malloc(stats.s);
	prm = (struct lbdr_param *)stats.data;

	i = ioctl(lb_dev, LBDRSTATS, stats);
	if(i == -1) {
		printlog(0, "ioctl error: %s", strerror(errno));
	}

	for(i = 0; i < nb_srv; i++) {

	}

	free(stats.data);
	stats.data = NULL;

	return 0;
}
	*/


static char *get_day(int d)
{
	switch(d) {
		case 0: return "Sun"; break;
		case 1: return "Mon"; break;
		case 2: return "Tue"; break;
		case 3: return "Wed"; break;
		case 4: return "Thu"; break;
		case 5: return "Fri"; break;
		case 6: return "Sat"; break;
		default: break;
	}
	return "";
}

static char *get_month(int m)
{
	switch(m) {
		case 0: return "Jan"; break;
		case 1: return "Feb"; break;
		case 2: return "Mar"; break;
		case 3: return "Apr"; break;
		case 4: return "May"; break;
		case 5: return "Jun"; break;
		case 6: return "Jul"; break;
		case 7: return "Aug"; break;
		case 8: return "Sep"; break;
		case 9: return "Oct"; break;
		case 10: return "Nov"; break;
		case 11: return "Dec"; break;
		default: break;
	}
	return "";
}


static char * name_scheduler(int s)
{
	switch(s) {
		case DAEMON_SCHED_WRRL:  return "wrrl";
		case DAEMON_SCHED_WLCL:  return "wlcl";
		case LBDR_SCHED_WLC:     return "wlc";
		case LBDR_SCHED_WRR:     return "wrr";
		case LBDR_SCHED_LC:      return "lc";
		case LBDR_SCHED_RR:      return "rr";
		case DAEMON_SCHED_RD:    return "rd";
		default: break;
	}

	return "Unknown!";
}



static int choose_scheduler(char *s)
{
	if(!strncasecmp(s, "wrrl", 4))      return DAEMON_SCHED_WRRL;
	else if(!strncasecmp(s, "wlcl", 4)) return DAEMON_SCHED_WLCL;
	else if(!strncasecmp(s, "wlc", 3))  return LBDR_SCHED_WLC;
	else if(!strncasecmp(s, "wrr", 3))  return LBDR_SCHED_WRR;
	else if(!strncasecmp(s, "lc", 2))   return LBDR_SCHED_LC;
	else if(!strncasecmp(s, "rr", 2))   return LBDR_SCHED_RR;
	else if(!strncasecmp(s, "rd", 2))   return DAEMON_SCHED_RD;

	return -1;
}



static int register_config(int nb_ip, struct sockaddr_in **ip_dst, int port, int hashsize, int sched, int tls, char *rrdfile, char *rrdname, struct sockaddr_in *ip_srv, int weight, char *url, int con, int rcv, int rtr, int nb_rtr, int wai, int syn_to, int est_to)
{
	struct lb_server  *srv = NULL;
	struct lb_server  *lsrv = NULL;
	struct lb_service *svice = NULL;
	struct lb_service *lsvice = NULL;
	struct lb_service *nsvice = NULL;
	int i;

	srv = server_available;
	while(srv) {
		if(srv->ip->sin_addr.s_addr == ip_srv->sin_addr.s_addr)
			break;
		lsrv = srv;
		srv = srv->next;
	}

	if(!srv) {
		printlog(5, "Add server: %u.%u.%u.%u", ADDR(ip_srv->sin_addr.s_addr));

		srv = malloc(sizeof(struct lb_server));
		if(!srv) exit(EXIT_FAILURE);

		bzero(srv, sizeof(struct lb_server));

		srv->ip = malloc(sizeof(struct sockaddr_storage));
		if(!srv->ip) exit(EXIT_FAILURE);

		memcpy(srv->ip, ip_srv, sizeof(struct sockaddr_in));
		srv->ip->sin_port = htons(port);
		srv->status = 1;
		srv->next = NULL;
		srv->lbdr_service = NULL;
		srv->pid = 0;
		srv->load = 0;
		srv->error = 0;
		srv->conn_to = con;
		srv->recv_to = rcv;
		srv->retry_to = rtr;
		srv->nb_retry = nb_rtr;
		srv->wait_to = wai;
		if(url) {
			srv->file = malloc(strlen(url)+2);
			snprintf(srv->file, strlen(url)+1, "%s", url);
		} else {
			srv->file = malloc(MAXNAMESIZE);
			snprintf(srv->file, MAXNAMESIZE, "index.html");
		}

		if(!server_available) {
			server_available = srv;
		} else {
			lsrv->next = srv;
		}
	} else {
		printlog(5, "Already defined server: %u.%u.%u.%u", ADDR(ip_srv->sin_addr.s_addr));
	}

	lsvice = srv->lbdr_service;
	if(lsvice) {
		while(lsvice->next) {
			lsvice = lsvice->next;
		}
	}

	nsvice = NULL;
	for(i = 0; i < nb_ip; i++) {
		svice = malloc(sizeof(struct lb_service));
		bzero(svice, sizeof(struct lb_service));
		if(!svice) exit(EXIT_FAILURE);
		svice->ip = malloc(sizeof(struct sockaddr_storage));
		svice->ip_srv = malloc(sizeof(struct sockaddr_storage));
		if(!svice->ip || !svice->ip_srv) exit(EXIT_FAILURE);

		svice->next = NULL;
		svice->weight = weight;	
		svice->tls = tls;
		svice->sched = sched;
		svice->syn_to = syn_to;
		svice->est_to = est_to;
		svice->hashsize = hashsize;
		if (rrdfile && rrdname) {
			svice->rrdname = malloc(strlen(rrdname)+1); 
			strlcpy(svice->rrdname, rrdname, strlen(rrdname));
		}
		if(use_clone && i > 0) {
			svice->cloned = 1;
			memcpy(svice->ip, ip_dst[0], sizeof(struct sockaddr_in));
			memcpy(svice->ip_srv, ip_dst[i], sizeof(struct sockaddr_in));

		} else {
			svice->cloned = 0;
			memcpy(svice->ip, ip_dst[i], sizeof(struct sockaddr_in));
			memcpy(svice->ip_srv, ip_srv, sizeof(struct sockaddr_in));
		}
		svice->ip->sin_port = htons(port);
		svice->ip_srv->sin_port = htons(port);
		printlog(5, "Add %s: %u.%u.%u.%u:%i %s %s (H: %i, RRD: %s@%s)", svice->cloned ? "clone  " : "service", ADDR(svice->ip->sin_addr.s_addr), ntohs(svice->ip->sin_port), name_scheduler(svice->sched), svice->tls ? "SSL" : "", svice->hashsize, svice->rrdname, svice->rrdfile);
		if(!lsvice) srv->lbdr_service = svice;
		else lsvice->next = svice;
		lsvice = svice;
	}

	return 0;
}

int get_int(char *word, int def)
{
	int i;

	for(i = 0; i < (strlen(word)-1); i++) {
		if(!isdigit((int)word[i])) {
			free(word);
			printf("'%s' is not a number\n", word);
			return def;
		}
	}

	i = atoi(word);
	free(word);

	return i;
}

static struct sockaddr_in *read_ip(FILE *fd)
{
	struct sockaddr_in *ip;
	char *ptr;

	ip = malloc(sizeof(struct sockaddr_in));
	ptr = read_word(fd);

	/*
	port = strchr(ptr, ':');

	if(!port) {
		printlog(3, "Using default port of 80 for service: %s\n", ptr);
		ip->sin_port = htons(80);
	} else {
		*port = 0;
		port++;
		ip->sin_port = htons(atoi(port));
	}
	*/
	if(is_not_ip_valid(ptr)) {
		printlog(1, "%s is not a valid IP\n", ptr);
		free(ip);
		free(ptr);
		return NULL;
	}

	ip->sin_family = AF_INET;
	ip->sin_addr.s_addr = inet_addr(ptr);

	free(ptr);

	return ip;
}

static struct sockaddr_in **read_mip(FILE *fd, int *nb)
{
	struct sockaddr_in **ip = NULL;
	char *ptr;
	int i = 0;

	ptr = read_word(fd);

	if(ptr[0] == '(' || ptr[0] == '{') {
		ptr = read_word(fd);
		do {
			i++;
			ip = realloc(ip, sizeof(struct sockaddr_in *) * i);
			ip[i-1] = malloc(sizeof(struct sockaddr_in));

			/*
			port = strchr(ptr, ':');

			if(!port) {
				printlog(3, "Using default port of 80 for service: %s\n", ptr);
				ip[i-1]->sin_port = htons(80);
			} else {
				*port = 0;
				port++;
				ip[i-1]->sin_port = htons(atoi(port));
			}
			*/

			if(is_not_ip_valid(ptr)) {
				printlog(1, "%s is not a valid IP\n", ptr);
				free(ip[i-1]);
				free(ptr);
				i--;
			}

			ip[i-1]->sin_family = AF_INET;
			ip[i-1]->sin_addr.s_addr = inet_addr(ptr);
			free(ptr);
			ptr = read_word(fd);
			if(!ptr) break;
		} while(ptr[0] != ')');
	} else {
		ip = malloc(sizeof(struct sockaddr_in *));
		ip[0] = malloc(sizeof(struct sockaddr_in));

		/*
		port = strchr(ptr, ':');

		if(!port) {
			printlog(3, "Using default port of 80 for service: %s\n", ptr);
			ip[0]->sin_port = htons(80);
		} else {
			*port = 0;
			port++;
			ip[0]->sin_port = htons(atoi(port));
		}
		*/

		if(is_not_ip_valid(ptr)) {
			printlog(1, "%s is not a valid IP\n", ptr);
			free(ip[0]);
			free(ip);
			free(ptr);
			return NULL;
		}

		i = 1;
		ip[0]->sin_family = AF_INET;
		ip[0]->sin_addr.s_addr = inet_addr(ptr);

		free(ptr);
	}

	*nb = i;

	return ip;
}


static int send_socket_data(int sd, char *code, char *s, ...)
{
	char buf[MAXBUFSIZE];
	va_list ap;
	int r;

	bzero(buf, MAXBUFSIZE);
	va_start(ap, s);
	vsnprintf(buf, MAXBUFSIZE, s, ap);
	va_end(ap);

	if(strlen(buf) > 0) {
		printlog(5, "SEND: %s", buf);
		if(send(sd, buf, strlen(buf), 0) == -1) {
			printlog(4, "send error: %s", strerror(errno));
			return -1;
		}
		bzero(buf, MAXBUFSIZE);
	}

	if(code && (strlen(code) >= 3)) {
		r = recv(sd, buf, MAXBUFSIZE, 0);
		printlog(5, "RECV: %s", buf);

		if(r == -1) {
			printlog(4, "recv error: %s", strerror(errno));
			return -1;
		}
		if(buf[0] != code[0] || buf[1] != code[1] || buf[2] != code[2]) {
			printlog(4, "mail protocol error");
			return -1;
		}
	}

	return 0;
}




static int send_email(char *msg, char *subj, ...)
{
	struct sockaddr_in ip;
	struct tm *ti;
	char subject[MAXBUFSIZE];
	va_list ap;
	time_t t;
	int sd;

	if(!email || !mailserver) return 0;

	ip.sin_family = AF_INET;
	ip.sin_addr.s_addr = inet_addr(mailserver);
	ip.sin_port = htons(25);

	bzero(subject, MAXBUFSIZE);
	va_start(ap, subj);
	vsnprintf(subject, MAXBUFSIZE, subj, ap);
	va_end(ap);

	if((sd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		printlog(3, "Can't open TCP socket : %s", strerror(errno));
		printlog(1, "Unable to send email");
		return -1;
	}

        if(connect(sd, (struct sockaddr *)&ip, sizeof(struct sockaddr_in)) == -1) {
                printlog(3, "Can't connect to %s:%i: %s\n", inet_ntoa(ip.sin_addr),
                	ntohs(ip.sin_port), strerror(errno));
		printlog(1, "Unable to send email..");
                return -1;
        }

	printlog(4, "connected to TCP %u.%u.%u.%u:%i", ADDR(ip.sin_addr.s_addr), ntohs(ip.sin_port));

	t = time(NULL);
	ti = localtime(&t);
	if(send_socket_data(sd, "220", "")) goto end;
	if(send_socket_data(sd, "250", "helo moto\r\n")) goto end;
	if(send_socket_data(sd, "250", "mail from: drlb@%s\r\n", hostname)) goto end;
	if(send_socket_data(sd, "250", "rcpt to: %s\r\n", email)) goto end;
	if(send_socket_data(sd, "354", "data\r\n")) goto end;
	if(send_socket_data(sd, NULL, "Date: %s, %i %s %i %i:%i:%i %c%i\r\n",
					get_day(ti->tm_wday), ti->tm_mday, get_month(ti->tm_mon),
					ti->tm_year + 1900, ti->tm_hour, ti->tm_min, ti->tm_sec,
					ti->tm_gmtoff > 0 ? '+' : '-', abs((int)ti->tm_gmtoff/36))) goto end;
	if(send_socket_data(sd, NULL, "From: drlb@%s\r\n", hostname)) goto end;
	if(send_socket_data(sd, NULL, "To: %s\r\n", email)) goto end;
	if(send_socket_data(sd, NULL, "Subject: %s\r\n", subject)) goto end;
	if(send_socket_data(sd, NULL, "Mime-Version: 1.0\r\n")) goto end;
	if(send_socket_data(sd, NULL, "Content-Type: text/plain; charset=iso-8859-15\r\n")) goto end;
	if(send_socket_data(sd, NULL, "Content-Disposition: inline\r\n")) goto end;
	if(send_socket_data(sd, "250", "\r\n%s\r\n.\r\n", msg)) goto end;

end:
	send_socket_data(sd, "221", "quit\r\n");
	close(sd);

	return 0;
}


static size_t into_memory(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	struct data *mem = (struct data *)data;

	mem->memory = (char *)realloc(mem->memory, mem->size + realsize + 1);
	if (mem->memory) {
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}
	return realsize;
}

static void chg_active(struct lb_service *service, int status)
{
	struct lb_service *tmp = service;

	while(tmp) {
		if(!tmp->cloned) {
			lb_ioctl(tmp->ip, tmp->ip_srv, LBDRCHGSRV, status, tmp->sched, tmp->weight, 0, tmp->hashsize);
		}
		tmp = tmp->next;
	}
}


char *get_downtime(int sec)
{
	char *u;
	int h = 0;
	int d = 0;
	int m = 0;
	int s = 0;

	d = abs(sec / 86400);
	h = abs((sec - d*86400) / 3600);
	m = abs((sec - d*86400 - h*3600) / 60);
	s = sec - d*86400 - h*3600 - m*60;

	u = malloc(MAXNAMESIZE);
	snprintf(u, MAXNAMESIZE, "%i days, %02i:%02i:%02i", d, h, m, s);

	return u;
}

#ifdef USE_CURL
static void print_http_info(struct in_addr in, CURL *curl)
{
	long l;
	double d;
	double t;

	if(curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &l) != CURLE_OK) {
		l = -1;
	}
	if(curl_easy_getinfo(curl, CURLINFO_TOTAL_TIME, &t) != CURLE_OK) {
		t = -1;
	}
	if(curl_easy_getinfo(curl, CURLINFO_SIZE_DOWNLOAD, &d) != CURLE_OK) {
		d = -1;
	}

	printlog(4, "%u.%u.%u.%u Response: %li (%.0f bytes in %.3f secs)", ADDR(in.s_addr), l, d, t);
}

static void cleanup_curl(void *arg)
{
	CURL *c = (CURL *)arg;

	curl_easy_cleanup(c);
}

static void * monitoring(void *arg)
{
	struct lb_server *srv = (struct lb_server *)arg;
	struct data data;
	CURL *curl;
	CURLcode res = 0;
	char url[MAXBUFSIZE];
	char *buf, *up = NULL;

	snprintf(url, MAXBUFSIZE,  "http://%u.%u.%u.%u/%s", ADDR(srv->ip->sin_addr.s_addr), srv->file ? srv->file : "load.php");

	data.memory = NULL;
	data.size = 0;

	srv->pid = pthread_self();
	curl = curl_easy_init();
	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_FAILONERROR, 1);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, into_memory);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&data);
	curl_easy_setopt(curl, CURLOPT_USERAGENT, useragent);
	curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, srv->recv_to);
	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, srv->conn_to);
	if(debug == 2) curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);

	pthread_cleanup_push(cleanup_curl, (void *)curl);

	for(;;) {
		pthread_testcancel();
		res = 0;
		srv->error = 0;
		do {
			if(res) {
				srv->error++;
				sleep(srv->retry_to);
				if(srv->status) printlog(3, "%u.%u.%u.%u Error: %s. Retry %i...",
					ADDR(srv->ip->sin_addr.s_addr),
					curl_easy_strerror(res), srv->error);
			}
			
			res = curl_easy_perform(curl);
			if(srv->status == 1 && res == 7) break; /* can't con, don't retry */
		} while(res && (srv->error < srv->nb_retry));


		if(res) {
			if(srv->status == 1) {
				srv->total_errors++;
				chg_active(srv->lbdr_service, 0);
				printlog(2, "Set %u.%u.%u.%u Inactive", ADDR(srv->ip->sin_addr.s_addr));
				buf = malloc(MAXBUFSIZE);
				if(buf) {
					snprintf(buf, MAXBUFSIZE-1, "Error: %s\n\n"
							    "URL: http://%u.%u.%u.%u/%s\n"
							    "Number of inhibit:        %i times\n\n"
							    "Current settings are:\n"
							    "Timeout connect:          %i sec\n"
							    "Timeout receive:          %i sec\n"
							    "Timeout before retry:     %i sec\n"
							    "Number of retry:          %i sec\n"
							    "Timeout between two poll: %i sec\n"
							    , curl_easy_strerror(res),
							    ADDR(srv->ip->sin_addr.s_addr), srv->file,
							    srv->total_errors,
							    srv->conn_to, srv->recv_to, srv->retry_to,
							    srv->nb_retry, srv->wait_to);
					send_email(buf, "Set %u.%u.%u.%u Inactive", ADDR(srv->ip->sin_addr.s_addr));
					free(buf);
					buf = NULL;
				}
				srv->status = 0;
				srv->downtime = time(NULL);
			}
		} else {
			if(loglevel > 3) print_http_info(srv->ip->sin_addr, curl);
			if(srv->status == 0) {
				chg_active(srv->lbdr_service, 1);
				printlog(2, "Set %u.%u.%u.%u Active", ADDR(srv->ip->sin_addr.s_addr));
				srv->downtime = time(NULL) - srv->downtime;
				srv->total_downtime += srv->downtime;
				buf = malloc(MAXBUFSIZE);
				if(buf) {
					up = get_downtime(srv->total_downtime);
					snprintf(buf, MAXBUFSIZE-1, "HTTP response 200 OK\n\n"
							    "URL: http://%u.%u.%u.%u/%s\n"
							    "Downtime:		       %i sec\n"
							    "Total downtime:	       %s\n"
							    ,
							    ADDR(srv->ip->sin_addr.s_addr), srv->file,
								srv->downtime, up);
					send_email(buf, "Set %u.%u.%u.%u Active", ADDR(srv->ip->sin_addr.s_addr));
					free(buf);
					free(up);
					buf = NULL;
				}
				srv->status = 1;
			}
			/* TODO: parse result and extract load */
		}
		free(data.memory);
		data.memory = NULL;
		data.size = 0;

		sleep(srv->wait_to);
	}
		
	curl_easy_cleanup(curl);

	return NULL;
}
#else
static void * monitoring(void *arg)
{
	send_email(buf, "", 0);
	chg_active();
	return NULL;
}
#endif

static void set_timeout(struct sockaddr_in *ip, int syn, int est, int clo)
{
	struct lbdr_timeout to;
	int r;

	to.nt_syn = syn;
	to.nt_est = est;
	to.nt_clo = clo;
	memcpy(&(to.ip_dst), ip, sizeof(struct sockaddr_in));

	r = ioctl(lb_dev, LBDRCHGTO, &to);
	if(r == -1) {
		printlog(1, "lbdr_set_timeout(): %s", strerror(errno));
	}
}

int lb_add(struct sockaddr_in *svi, struct sockaddr_in *srv, int a, int s, int w, int p, int h, int c)
{
	struct lbdr_param *prm;
	struct sockaddr_in *d, *e;
	int i;

	prm = (struct lbdr_param *)stats.data;

	for(i = 0; i < info.nb_srv; i++) {
		d = (struct sockaddr_in *)&(prm[i].ip_dst);
		e = (struct sockaddr_in *)&(prm[i].ip_srv);
		if(d->sin_addr.s_addr == svi->sin_addr.s_addr &&
		   e->sin_addr.s_addr == srv->sin_addr.s_addr) {
		   	if((c == 0) && (prm[i].weight != w)) {
				lb_ioctl(svi, srv, LBDRCHGSRV, 1, s, w, p, h);
				prm[i].status = 0x03;
			}
			return 0;
		   }
	}

	if(c == 1) {
		lb_ioctl(svi, srv, LBDRCLONE, 1, s, w, p, h);
	} else {
		lb_ioctl(svi, srv, LBDRADDSRV, 1, s, w, p, h);
	}

	return 0;
}
	
static void populous(void)
{
	struct lb_service *service;
	struct lb_server *srv;
	int tmp_sched;
	int nb_srv = 0;

	srv = server_available;

	while(srv) {

		service = srv->lbdr_service;
		while(service) {
			if(service->sched == DAEMON_SCHED_RD) {
				/*listen_thread(service);*/
				service = service->next;
				continue;
			} else if(service->sched == DAEMON_SCHED_WRRL) tmp_sched = LBDR_SCHED_WRR;
			else if(service->sched == DAEMON_SCHED_WLCL) tmp_sched = LBDR_SCHED_WLC;
			else tmp_sched = service->sched;
			lb_add(service->ip, service->ip_srv, 1, tmp_sched, service->weight, service->tls, service->hashsize, service->cloned);
			/* TODO: do not set if already set */
			set_timeout(service->ip, service->syn_to, service->est_to, def_clo_to);
			service = service->next;
		}

		if(pthread_create (&(srv->pid), NULL, monitoring, (void *)srv) != 0) {
			printlog(0, "Can't create pthread: %s\n", strerror(errno));
		} else {
			printlog(3, "Start thread for %u.%u.%u.%u", ADDR(srv->ip->sin_addr.s_addr));
			nb_srv++;
		}

		srv = srv->next;
	}

	printlog(2, "Start monitoring %i servers.", nb_srv);

}

void armageddon(void)
{
	struct lb_server *last, *srv = server_available;
	struct lb_service *s, *ls;
	void *ret;

	while(srv) {
		if(pthread_cancel (srv->pid)) {
			printlog(0, "Can't cancel pthread: %s\n", strerror(errno));
		} else printlog(3, "Stop thread for %u.%u.%u.%u", ADDR(srv->ip->sin_addr.s_addr));
		srv = srv->next;
	}

	while(srv) {
		pthread_join(srv->pid, &ret);
		last = srv;
		srv = srv->next;
		s = last->lbdr_service;
		while(s) {
			ls = s;
			s = s->next;
			free(ls->ip);
			free(ls);
		}
		free(last->ip);
		/*free(last->url);*/
		free(last);
	}

	server_available = NULL;
}



void get_lb_status(void)
{
	int r;

	r = ioctl(lb_dev, LBDRINFO, &info);
	if(r == -1) {
		perror("lbdr");
	}

	stats.size = info.nb_srv * sizeof(struct lbdr_param);
	stats.data = malloc(stats.size);

	r = ioctl(lb_dev, LBDRSTATS, &stats);

	if(r == -1) {
		perror("lbdr");
	}
}

void read_conf_service(FILE *fd)
{
	char *word, *value;
	struct sockaddr_in **ip_dst = NULL;
	struct sockaddr_in *ip_srv = NULL;
	char *rrdfile = NULL, *rrdname = NULL, *url = NULL;
	int sched = LBDR_SCHED_RR, hashsize = 15, nb_ipservice = 0, persistence = 0, port = 80, weight = 10;
	int conn_to = def_conn_to, recv_to = def_recv_to, retry_to = def_retry_to, nb_retry = def_nb_retry, wait_to = def_wait_to, syn_to = def_syn_to, est_to = def_est_to;
	int configured = 0;

	while(!feof(fd)) {

		word = read_spaceword(fd);

		if(!word) continue;

newword:
		if(!strncasecmp(word, "{ ", 2)) {
			continue;
		} else if(!strncasecmp(word, "} ", 2)) {
			return;
		} else if(!strncasecmp(word, "server ", 7)) {
			configured = 1;
			ip_srv = read_ip(fd);
			word = read_spaceword(fd);
			if(strncasecmp(word, "weight ", 7)) {
				register_config(nb_ipservice, ip_dst, port, hashsize, sched, persistence,
						rrdfile, rrdname, ip_srv, weight, url,
						conn_to, recv_to, retry_to, nb_retry, wait_to, syn_to, est_to);
				goto newword;
			}
			weight = get_int(read_word(fd), 10);
			register_config(nb_ipservice, ip_dst, port, hashsize, sched, persistence,
					rrdfile, rrdname, ip_srv, weight, url,
					conn_to, recv_to, retry_to, nb_retry, wait_to, syn_to, est_to);
			free(word);
			continue;
		}

		if(configured) {
			printlog(1, "Ignoring keyword '%s' (must be before server's IP)", word);
		}

		if(!strncasecmp(word, "ip ", 3)) {
			if(ip_dst) {
				printf("IP already defined for service\n");
				continue;
			}
			ip_dst = read_mip(fd, &(nb_ipservice));
		} else if(!strncasecmp(word, "scheduler ", 10)) {
			value = read_word(fd);
			sched = choose_scheduler(value);
			if(sched == -1) {
				printlog(3, "Unknow scheduler: %s  Default to RR.\n", value);
				sched = LBDR_SCHED_RR;
			}
			free(value);
		} else if(!strncasecmp(word, "hashsize ", 9)) {
			hashsize = get_int(read_word(fd), hashsize);
		} else if(!strncasecmp(word, "url ", 4)) {
			url = read_word(fd);
		} else if(!strncasecmp(word, "rrdname ", 8)) {
			rrdname = read_word(fd);
		} else if(!strncasecmp(word, "rrdfile ", 8)) {
			rrdfile = read_word(fd);
		} else if(!strncasecmp(word, "port ", 5)) {
			port = get_int(read_word(fd), port);
		} else if(!(strncasecmp(word, "tls ", 4) && strncasecmp(word, "ssl ", 4))) {
			persistence = 1;
		} else if(!strncasecmp(word, "get_timeout_retry ", 18)) {
			retry_to = get_int(read_word(fd), def_retry_to);
		} else if(!strncasecmp(word, "get_timeout_conn ", 17)) {
			conn_to = get_int(read_word(fd), def_conn_to);
		} else if(!strncasecmp(word, "get_timeout_recv ", 17)) {
			recv_to = get_int(read_word(fd), def_recv_to);
		} else if(!strncasecmp(word, "lbdr_timeout_syn ", 17)) {
			syn_to = get_int(read_word(fd), def_syn_to);
		} else if(!strncasecmp(word, "lbdr_timeout_est ", 17)) {
			est_to = get_int(read_word(fd), def_est_to);
		} else if(!strncasecmp(word, "timeout_wait ", 13)) {
			wait_to = get_int(read_word(fd), def_wait_to);
		} else if(!strncasecmp(word, "get_retry ", 10)) {
			nb_retry = get_int(read_word(fd), def_nb_retry);
		} else {
			printf("Keyword '%s' not understood for service configuration\n", word);
		}
		free(word);
	}

	return;
}

static void read_config(char *file)
{
	FILE *fd;
	char *word;

        fd = fopen(file, "r");

        if(fd == NULL) {
                printlog(0, "Can't open %s: %s\n", file, strerror(errno));
                exit(EXIT_FAILURE);
        }

	while(!feof(fd)) {
static int def_conn_to = 1;
static int def_recv_to = 2;
static int def_retry_to = 3;
static int def_nb_retry = 3;
static int def_wait_to = 10;


		word = read_spaceword(fd);

		if(!word) continue;

		if(!strncasecmp(word, "get_timeout_retry ", 18)) {
			def_retry_to = get_int(read_word(fd), def_retry_to);
		} else if(!strncasecmp(word, "get_timeout_conn ", 17)) {
			def_conn_to = get_int(read_word(fd), def_conn_to);
		} else if(!strncasecmp(word, "get_timeout_recv ", 17)) {
			def_recv_to = get_int(read_word(fd), def_recv_to);
		} else if(!strncasecmp(word, "lbdr_timeout_syn ", 17)) {
			def_syn_to = get_int(read_word(fd), def_syn_to);
		} else if(!strncasecmp(word, "lbdr_timeout_est ", 17)) {
			def_est_to = get_int(read_word(fd), def_est_to);
		} else if(!strncasecmp(word, "lbdr_timeout_clo ", 17)) {
			def_clo_to = get_int(read_word(fd), def_clo_to);
		} else if(!strncasecmp(word, "timeout_wait ", 13)) {
			def_wait_to = get_int(read_word(fd), def_wait_to);
		} else if(!strncasecmp(word, "get_retry ", 10)) {
			def_nb_retry = get_int(read_word(fd), def_nb_retry);
		} else if(!strncasecmp(word, "useragent ", 10)) {
			if(useragent) printlog(2, "useragent already defined");
			useragent = read_word(fd);
		} else if(!strncasecmp(word, "loglevel ", 9)) {
			loglevel = get_int(read_word(fd), loglevel);
		} else if(!strncasecmp(word, "pidfile ", 8)) {
			if(pid_file) printlog(2, "pidfile already defined");
			pid_file = read_word(fd);
		} else if(!strncasecmp(word, "service ", 8)) {
			read_conf_service(fd);
		} else if(!strncasecmp(word, "alert ", 6)) {
			if(email) printlog(2, "alert alert defined");
			email = read_word(fd);
		} else if(!strncasecmp(word, "group ", 6)) {
			if(group) printlog(2, "group already defined");
			group = read_word(fd);
		} else if(!strncasecmp(word, "clone ", 6)) {
			use_clone = 1;
		} else if(!strncasecmp(word, "user ", 5)) {
			if(user) printlog(2, "user already defined");
			user = read_word(fd);
		} else if(!strncasecmp(word, "smtp ", 5)) {
			if(mailserver) printlog(2, "smtp already defined");
			mailserver = read_word(fd);
		} else {
			printlog(3, "keyword '%s' not understood.", word);
		}
		free(word);
	}

	fclose(fd);
}

static void sig_hup(int s)
{
	printlog(3, "Rereading config file");
	reread = 1;
	process = 0;
}

static void sig_int(int s)
{
	process = 0;
}

int main(int ac, char *av[])
{
	struct passwd *pw = NULL;
	struct group *gr = NULL;
	char *config_file = NULL;
	char c;

	config_file = malloc(MAXNAMESIZE);
	strlcpy(config_file, "lbdyn.conf", MAXNAMESIZE);

	while((c = getopt(ac, av, "Dhdf:")) != -1) {
		switch(c) {
		case 'f':
			strlcpy(config_file, optarg, MAXNAMESIZE);
			break;
		case 'D':
			debug++;
		case 'd':
			debug++;
			use_syslog = 0;
			break;
		case 'h':
		default:
			printf( "Usage: %s -d -f <file>\n"
				"\t-d\tDon't fork. Be verbose.\n"
				, av[0]);
			return EXIT_FAILURE;
			break;
		}
	}

	ac -= optind;
	av += optind;

	read_config(config_file);

	if(!useragent) {
		useragent = malloc(MAXNAMESIZE);
		strlcpy(useragent, "lbdyn", MAXNAMESIZE);
	}

	if(debug) {
		if(loglevel < 3) loglevel = 3;
		if(debug == 2) loglevel = 5;
	} else {
		daemon(1, 0);
	}

	if(pid_file) {
		if(create_pid(pid_file)) {
			/* free(all) */
			return -1;
		}
		printlog(3, "Pidfile: %s", pid_file);
	}

	hostname = malloc(MAXNAMESIZE);
	gethostname(hostname, MAXNAMESIZE);

	mtx_ioctl = malloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(mtx_ioctl, NULL);
#ifdef USE_CURL
	curl_global_init(CURL_GLOBAL_ALL);
#endif

	lb_dev = open(LBDR_DEVICE, O_RDWR);
	if(lb_dev == -1) {
		printlog(0, "Can't open %s : %s", LBDR_DEVICE, strerror(errno));
		/* free(all) */
		return -1;
	}

	signal(SIGHUP, sig_hup);
	signal(SIGINT, sig_int);

reading:
	reread = 0;
	process = 1;

	if(use_syslog)
		openlog(useragent, LOG_PID, LOG_USER);

	if(user) {
		pw = getpwnam(user);
		setuid(pw->pw_uid);
		if(group) {
			gr = getgrnam(group);
			setgid(gr->gr_gid);
		}
		printlog(3, "Change UID to %s(%i):%s(%i)", user, pw->pw_uid, group, gr->gr_gid);
	}

	if(email && mailserver)
		printlog(2, "Send alert to: %s via %s", email, mailserver);

	get_lb_status();

	populous();

	/* delete() server not in conf */

	for(; process == 1 ; sleep(1))  ;
		/*update_status()*/

	armageddon();

	closelog();

	free(hostname); hostname = NULL;
	free(user); user = NULL;
	free(group); group = NULL;
	free(mailserver); mailserver = NULL;
	free(email); email = NULL;
	free(pid_file); pid_file = NULL;
	free(useragent); useragent = NULL;
	free(stats.data);
		
	if(reread) {
		read_config(config_file);
		goto reading;
	}

	close(lb_dev);
	pthread_mutex_destroy(mtx_ioctl);
#ifdef USE_CURL
	curl_global_cleanup();
#endif
	if(pid_file) remove_pid();

	free(mtx_ioctl);
	free(config_file);

	return 0;
}

/* EOF */

